--Copyright ©2020, Wirepath Home Systems, LLC. All rights reserved.

__LOG_NAME__ = "EA-MINI-51D200"

require "law.baseDriver"
require "law.objects"
require "law.SystemEventManager"
require "law.ServiceLocator"
require "law.TimerManager"
require "law.StringBuilder"
require "law.Log"
require "law.VariableManager"
require "law.Debugger"
require "law.Xml"


JSON=require "json"

require "epi_protocol"
require "epi_presets"
require "epi_data"

DEFAULT_PROXY_BINDING_ID = 5001

MAX_VOL_VALUE = 100
MIN_VOL_VALUE = 0
MAX_TREBLE_VALUE = 12.0
MIN_TREBLE_VALUE = -12.0
MAX_BASS_VALUE = 12.0
MIN_BASS_VALUE = -12.0
MAX_BALANCE_VALUE = 12.0
MIN_BALANCE_VALUE = -12.0

DE_BINDING_ADDED = 15
DE_NETWORK_BINDING_REMOVED = 16

DV_ACTIVE_ZONE_NAME = "ActiveZone"

AUTO_ARC_INPUT_SELECTID = 7
g_useAutoArcSelect = true
 
DEFAULT_BINDING_VOLUME_ROOM = 20
DEFAULT_BINDING_VOLUME_NON_ROOM = 100

g_startTime = socket.gettime()
g_lateInited = false
g_shouldSuppress = false

ON_PROPERTY_CHANGED = {}

NOTIFY_OUTPUT_INDEX = 4000

-- index used to track which output the Composer Property page is using
g_properyOutputIndexSelected = 1
-- index used to track which input the Composer Property page is using
g_properyInputIndexSelected = 1
g_needLoginRetry = false

-- current state of volume and selection
g_currentInputSelected = AUTO_ARC_INPUT_SELECTID --setting this to arc for auto-arc selection
g_currentVolume = 0
g_currentMuted = false

g_roomBindingNames = {}
g_volumeOutputIsBoundMap = {}
g_delayedInputCmdsTimers = {}
g_unparsedJsonBuffer = ""
--------------------------------------------------------------------------------------------
--							DRIVER SETUP / MAINTENANCE METHODS			 				  
--------------------------------------------------------------------------------------------

function SendProxyAllNotifyData()
	NotifyVolumeLevel(NOTIFY_OUTPUT_INDEX)
	SendNotify("MUTE_CHANGED", {MUTE = g_currentMuted}, NOTIFY_OUTPUT_INDEX)
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, 'INPUT_OUTPUT_CHANGED', {INPUT = -1, OUTPUT = NOTIFY_OUTPUT_INDEX, AUDIO=true, VIDEO=false}, "NOTIFY")  
    SendNotifyInput("SURROUND_MODE_CHANGED",  {SURROUND_MODE = tostring(g_epiData.inputs[g_currentInputSelected].soundMode)}, NOTIFY_OUTPUT_INDEX)
    for i = 1, g_epiData.inputCount do
		SendNotifyInput("INPUT_GAIN_CHANGED",  {LEVEL = g_epiData.inputs[i]}, i)
	end
end

function OnDriverInit()
	print("OnDriverInit")	
	SystemEventManager:new(true)	
	VariableManager:new(true)
	TimerManager:new(true)			
	g_lateInited = false
	C4:RegisterSystemEvent(DE_BINDING_ADDED, 0) 
	Log:OutputC4Log(false)
	Log:OutputPrint(false)
	Log:SetLogLevel(2)
	local outputCount = tonumber(C4:GetCapability("audio_provider_count"))
	local inputCount = tonumber(C4:GetCapability("audio_consumer_count")) + tonumber(C4:GetCapability("video_consumer_count"))
	-- note: I am using one based index for inputs and outputs, so a zero will mean not connected
    if ((g_epiData == nil) or (g_epiData.outputs == nil)) then
		-- setup data object
		g_epiData = epi_data.Init(outputCount, inputCount)
	end
    epi_protocol.init(outputCount, inputCount)
	g_connected = false
	g_shouldSuppress = false
	g_needToSync = true
    g_useAutoArcSelect = true
	g_delayedInputCmdsTimers = {}
end

function OnDriverLateInit()
	print("OnDriverLateInit")
	g_lateInited = true
	g_shouldSuppress = false
	SendProxyAllNotifyData()
    ON_PROPERTY_CHANGED.InputIndex("1")
    ON_PROPERTY_CHANGED.OutputIndex("1")
    C4:AddVariable(DV_ACTIVE_ZONE_NAME, 0, "BOOL")

end

function OnDriverDestroyed()
    print("OnDriverDestroyed")
    ServiceLocator.TimerManager:destroy()
    ServiceLocator.VariableManager:destroy()
    ServiceLocator.SystemEventManager:destroy()
    PersistData = g_epiData
    C4:InvalidateState() 
end

function OnConnectionStatusChanged(idBinding, nPort, strStatus)
	if (idBinding == 6001) then
		if (strStatus == "ONLINE") then
			Log:Warn("TCP connection established to device")
			g_connected = true
            g_unparsedJsonBuffer = ""
			g_needToSync = true
            --epi_protocol.setPowerOn()
			epi_protocol.login(g_epiData.userName, g_epiData.password)
            g_errorCount = 0
            g_okCount = 0
		else 
			Log:Warn("TCP connection was lost to device")
			g_connected = false
		end
        epi_protocol.ConnectionStatusChanged(g_connected)
        C4:UpdateProperty("Connection Status", strStatus)
	end
end

function OnPoll(bindingId, port)
    if (pollCount == nil) then pollCount = 0 end
    pollCount = pollCount + 1
    if (g_needLoginRetry) then
    	epi_protocol.login(g_epiData.userName, g_epiData.password)
        g_needLoginRetry = false
    end
    epi_protocol.getPowerStatus()
end

function OnSystemEvent(data)
	Log:Debug("System Event occurred - data: " .. data)
--	local xml = Xml:new(data)
end


------------------------ Handler for Properties ------------------------
 -- All property changes are initiated from composer 
 -- We must use the appropriate INPUT or OUTPUT index to know context for most property changes
function OnPropertyChanged(sProperty)
    printd("opc " .. sProperty)
	local propertyValue = Properties[sProperty]
    printd("opc val " .. tostring(propertyValue))
	printd("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		ON_PROPERTY_CHANGED[sProperty](propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		ON_PROPERTY_CHANGED[trimmedProperty](propertyValue)
	end
end

function ON_PROPERTY_CHANGED.DebugMode(propertyValue)
   if (g_printTimerId ~= nil) then g_printTimerId:Cancel() end
   g_printTimerId = C4:SetTimer(300000, LogTimerHandler)
	Log:OutputPrint(propertyValue:find("Print") ~= nil)
	Log:OutputC4Log(propertyValue:find("Log") ~= nil)
	if (propertyValue == "Off") then
		return
	end
end

function ON_PROPERTY_CHANGED.DebugLevel(propertyValue)
   Log:SetLogLevel(tonumber(string.sub(propertyValue, 1, 1)))
   if (g_printTimerId ~= nil) then g_printTimerId:Cancel() end
   g_printTimerId = C4:SetTimer(300000, LogTimerHandler)
end

function ON_PROPERTY_CHANGED.AmplifierName(propertyValue)
    g_epiData.ampName = propertyValue
    epi_protocol.setAmpName(g_epiData.ampName)
end

function ON_PROPERTY_CHANGED.UserName(propertyValue)
    g_epiData.userName = propertyValue
    if (g_epiData.password ~= "") then
        epi_protocol.login(g_epiData.userName, g_epiData.password)
    end
end

function ON_PROPERTY_CHANGED.MasterVolume(propertyValue)
    g_currentVolume = tonumber(propertyValue)
    epi_protocol.setVolume(g_currentInputSelected, g_currentVolume)
end

function ON_PROPERTY_CHANGED.AutoARCSelect(propertyValue)
    g_useAutoArcSelect = true
	if (propertyValue == "Off") then
        g_useAutoArcSelect = false
    end
end

function ON_PROPERTY_CHANGED.Password(propertyValue)
    g_epiData.password = propertyValue
    if (g_epiData.userName ~= "") then
        epi_protocol.login(g_epiData.userName, g_epiData.password)
    end
end

------------------------  Ip Address Properties ------------------------
function ON_PROPERTY_CHANGED.DHCP(propertyValue)
    g_epiData.DHCP = propertyValue
    local hideIpStuff = 0
    if (propertyValue == "on") then hideIpStuff = 1 end
    C4:SetPropertyAttribs("IP Address", hideIpStuff)
    C4:SetPropertyAttribs("IP Gateway Address", hideIpStuff)
    C4:SetPropertyAttribs("IP DNS Address", hideIpStuff)
    C4:SetPropertyAttribs("IP Subnet Mask", hideIpStuff)
end

function ON_PROPERTY_CHANGED.IPAddress(propertyValue)
    g_epiData.IPADDRESS = propertyValue
end

function ON_PROPERTY_CHANGED.IPSubnetMask(propertyValue)
    g_epiData.MASKADDRESS = propertyValue
end

function ON_PROPERTY_CHANGED.IPGatewayAddress(propertyValue)
    g_epiData.GATEWAYADDRESS = propertyValue
end

function ON_PROPERTY_CHANGED.IPDNSAddress(propertyValue)
    g_epiData.DNSADDRESS = propertyValue
end

function ON_PROPERTY_CHANGED.ApplyIPChanges(propertyValue)
    if (propertyValue == "Yes") then
        if (g_epiData.DHCP == "off") then
            -- validate static ip settings
            if ((g_epiData.IPADDRESS:find(".", 1, true) ~= nil) and
               (g_epiData.GATEWAYADDRESS:find(".", 1, true) ~= nil) and
               (g_epiData.DNSADDRESS:find(".", 1, true) ~= nil) and
               (g_epiData.MASKADDRESS:find(".", 1, true) ~= nil)) then
                epi_protocol.setIpAddress(g_epiData.DHCP, g_epiData.IPADDRESS, g_epiData.GATEWAYADDRESS, g_epiData.DNSADDRESS, g_epiData.MASKADDRESS) 
                print("Configured for static ip")
                C4:UpdateProperty("Web Config", "http://" .. g_epiData.IPADDRESS)
            else
                print("Invalid static ip address information entered")
            end
        else
            g_epiData.DHCP = "on"
            epi_protocol.setIpAddress(g_epiData.DHCP, g_epiData.IPADDRESS, g_epiData.GATEWAYADDRESS, g_epiData.DNSADDRESS, g_epiData.MASKADDRESS) 
            print("Configured for DHCP")
            C4:UpdateProperty("Web Config", "http://" .. g_epiData.IPADDRESS)
        end
        C4:UpdateProperty("Apply IP Changes", "----")
    end
end

------------------------  Input Properties ------------------------
function SendComposerUpdatedInputNames()
    newList = ""
	for i = 1, g_epiData.inputCount do
        if (i ~=g_epiData.inputCount) then
            newList = newList .. (i .. " - " .. g_epiData.inputs[i].name .. ",")
        else
            newList = newList .. (i .. " - " .. g_epiData.inputs[i].name)
        end
	end
    C4:UpdatePropertyList("Input Index", newList, tostring(g_properyInputIndexSelected .. " - " .. g_epiData.inputs[g_properyInputIndexSelected].name))
end

function ON_PROPERTY_CHANGED.InputIndex(propertyValue)
    if (propertyValue ~= "" and propertyValue ~= nil) then
        g_properyInputIndexSelected = tonumber(string.sub(propertyValue, 1, 1))
        C4:UpdateProperty("Input Name", tostring(g_epiData.inputs[g_properyInputIndexSelected].name))
        C4:UpdateProperty("Input Trim", tostring(g_epiData.inputs[g_properyInputIndexSelected].trim))
    end
end

function ON_PROPERTY_CHANGED.InputName(propertyValue)
    g_epiData.inputs[g_properyInputIndexSelected].name = propertyValue
    epi_protocol.setInputName(g_properyInputIndexSelected, propertyValue)
    SendComposerUpdatedInputNames()
end

function ON_PROPERTY_CHANGED.InputTrim(propertyValue)
    g_epiData.inputs[g_properyInputIndexSelected].trim = propertyValue
    epi_protocol.setInputTrimLevel(g_properyInputIndexSelected, propertyValue)
end



------------------------  Output Properties ------------------------
function SendComposerUpdatedOutputNames()
    newList = ""
	for i = 1, g_epiData.outputCount do
        if (i ~=g_epiData.outputCount) then
            newList = newList .. (i .. " - " .. g_epiData.outputs[i].name .. ",")
        else
            newList = newList .. (i .. " - " .. g_epiData.outputs[i].name)
        end
	end
    C4:UpdatePropertyList("Output Index", newList, tostring(g_properyOutputIndexSelected .. " - " .. g_epiData.outputs[g_properyOutputIndexSelected].name))
end


function ON_PROPERTY_CHANGED.OutputIndex(propertyValue)
    if (propertyValue ~= "" and propertyValue ~= nil) then
        g_properyOutputIndexSelected = tonumber(string.sub(propertyValue, 1, 1))
        C4:UpdateProperty("Output Name", g_epiData.outputs[g_properyOutputIndexSelected].name)
        C4:UpdateProperty("Bridge Mode", g_epiData.outputs[g_properyOutputIndexSelected].bridgeMode)
        C4:UpdateProperty("Stereo Mode", g_epiData.outputs[g_properyOutputIndexSelected].stereoMode)
        C4:UpdateProperty("Muted", g_currentMuted)
        C4:UpdateProperty("Master Volume", tostring(g_currentVolume))
    end
end

function ON_PROPERTY_CHANGED.OutputName(propertyValue)
    g_epiData.outputs[g_properyOutputIndexSelected].name = propertyValue
    epi_protocol.setOutputName(g_properyOutputIndexSelected, propertyValue)
    SendComposerUpdatedOutputNames()
end


function ON_PROPERTY_CHANGED.BridgeMode(propertyValue)
    g_epiData.outputs[g_properyOutputIndexSelected].bridgeMode = propertyValue
    epi_protocol.setOutputBridgeMode(g_properyOutputIndexSelected, propertyValue)
end

function ON_PROPERTY_CHANGED.StereoMode(propertyValue)
    g_epiData.outputs[g_properyOutputIndexSelected].stereoMode = propertyValue
    epi_protocol.setOutputStereoMode(g_properyOutputIndexSelected, propertyValue)
end

function MuteStateToValue(state)
    if (state == "true") then
      return "Muted"
    end
    if (state == "True") then
      return "Muted"
    end
    if (state == "TRUE") then
      return "Muted"
    end
    if (tostring(state) == "1") then
      return "Muted"
    end
    if (state == "Muted") then
      return "Muted"
    end
    return "Unmuted"
end
 
function ON_PROPERTY_CHANGED.Muted(propertyValue)
    g_currentMuted =MuteStateToValue(propertyValue)
    Mute(g_properyOutputIndexSelected, propertyValue)
end

function ON_PROPERTY_CHANGED.Volume(propertyValue)
    Volume(g_properyOutputIndexSelected,  tonumber(propertyValue))
end

-- attrib->http://lua-users.org/wiki/StringRecipes
function string.starts(String,Start)
   return string.sub(String,1,string.len(Start))==Start
end

function SyncStateToDevice()
     g_needToSync = false
     if (g_currentInputSelected > 0) then
   	     epi_protocol.setOutputToInput(i, g_currentInputSelected)
         epi_protocol.setVolume(g_currentInputSelected, g_currentVolume)
         epi_protocol.getInputSetting(g_currentInputSelected)
    end
    epi_protocol.getPowerStatus()
    epi_protocol.getIpAddress()
    epi_protocol.getSystemSetting()
    epi_protocol.getCecSetting()
    epi_protocol.getGeneralSetting()
    epi_protocol.getOutputSetting()
    epi_protocol.getVolumes()
end

function string:split(pat)
  pat = pat or '%s+'
  local st, g = 1, self:gmatch("()("..pat..")")
  local function getter(segs, seps, sep, cap1, ...)
    st = sep and seps + #sep
    return self:sub(segs, (seps or 0) - 1), cap1 or sep, ...
  end
  return function() if st then return getter(st, g()) end end
end

RECEIVED_FROM_NETWORK = {}

errorCode = {}
errorCode["200"] = "Success"
errorCode["300"] = "Repeat operation"
errorCode["400"] = "User or password error"
errorCode["401"] = "Illegal request" 
errorCode["402"] = "Please login first" 
errorCode["403"] = "The password is default and needs to be changed"
errorCode["405"] = "Account is signed out" 
errorCode["406"] = "Too many incorrect passwords, account locked"
errorCode["407"] = "Enter the same password as before, please reset" 
errorCode["408"] = "Password set successfully" 
errorCode["500"] = "An unknown error occurred on the server"
errorCode["600"] = "The import file does not exist" 
errorCode["601"] = "File md5 values do not match" 
errorCode["602"] = "File type mismatch" 
errorCode["700"] = "The machine is in standby mode"
errorCode["701"] = "The machine is in voltagetriggermode"
errorCode["702"] = "The machine is in audiomode" 
errorCode["800"] = "Update success" 
errorCode["801"] = "Update fail"

function RECEIVED_FROM_NETWORK.login(jrep)
    local ec = tostring(jrep["status"])
    print("EC = " .. ec)
    local estring = errorCode[ec]
    if (estring ~= nil) then C4:UpdateProperty("Login Status", estring) end
    if (ec == "200") then
        SyncStateToDevice()
    end
end

function RECEIVED_FROM_NETWORK.get_systemsetting(jrep)
    local ec = tostring(jrep["status"])
    if (ec == "200") then
         tdata = jrep["value"]
         g_epiData.ampName = tdata["ampName"]
         C4:UpdateProperty("Amplifier Name", g_epiData.ampName)
         g_epiData.macAddress = tdata["macAddress"]
         C4:UpdateProperty("MAC Address", g_epiData.macAddress)
         g_epiData.firmwareVersion = tdata["firmwareVersion"]
         g_epiData.model = tdata["model"]
   end
end

function RECEIVED_FROM_NETWORK.get_powerstatus(jrep)
    local ec = tostring(jrep["status"])
    if (ec == "200") then
         local powerStatus = tostring(jrep["value"])
         if (powerStatus == "0") then
            Log:Debug("Received power status from device")
            --epi_protocol.setPowerOn()
         end
    end
    if (ec == "402") then
    	epi_protocol.login(g_epiData.userName, g_epiData.password)
        g_needLoginRetry = false
    end
end

function RECEIVED_FROM_NETWORK.get_ip(jrep)
    local ec = tostring(jrep["status"])
    if (ec == "200") then
         tip = jrep["value"]
         g_epiData.IPADDRESS = tip["ipaddress"]
         C4:UpdateProperty("IP Address", g_epiData.IPADDRESS)
         C4:UpdateProperty("Web Config", "http://" .. g_epiData.IPADDRESS)
         g_epiData.GATEWAYADDRESS = tip["gateway"]
         C4:UpdateProperty("IP Gateway Address", g_epiData.GATEWAYADDRESS)
         g_epiData.MASKADDRESS = tip["subnetmask"]
         C4:UpdateProperty("IP Subnet Mask", g_epiData.MASKADDRESS)
         g_epiData.DNSADDRESS = tip["dnsserver"]
         C4:UpdateProperty("IP DNS Address", g_epiData.DNSADDRESS)
         local dhcpVal = "on"
         if (tip["dhcp"] == "1") then
            g_epiData.DHCP = true
         else
            g_epiData.DHCP = false
            dhcpVal = "off"
         end
         C4:UpdateProperty("DHCP", dhcpVal)
         ON_PROPERTY_CHANGED.DHCP(dhcpVal)
    end
end

function RECEIVED_FROM_NETWORK.get_inputsetting(jrep)
    local ec = tostring(jrep["status"])
    if (ec == "200") then
        local vt = jrep["value"]
        for k,v in pairs(vt) do
            local tb = v
            local entry = tb["index"]+1
            local inputGain = tonumber(tb["inputGain"])
            g_epiData.inputs[entry].inputGain = inputGain
            local enabled = tonumber(tb["inputEnable"])
            g_epiData.inputs[entry].enabled = enabled
            local soundMode = tonumber(tb["soundMode"]) + 1
            g_epiData.inputs[entry].soundMode = soundMode
            local maxVolume = tonumber(tb["maxVolume"])
            g_epiData.inputs[entry].maxVolume = maxVolume
            if (entry == g_properyInputIndexSelected) then
	      SendNotifyInput("SURROUND_MODE_CHANGED",  {SURROUND_MODE = tostring(smode)}, NOTIFY_OUTPUT_INDEX)
               --C4:UpdateProperty("Input Trim", vol)
            end
        end
    end
end

function RECEIVED_FROM_NETWORK.get_mastervolume(jrep)
    local ec = tostring(jrep["status"])
    if (ec == "200") then
        local vt = jrep["value"]
        for k,v in pairs(vt) do
            tb = v
            index = tonumber(tb["index"]) + 1
            if (index == g_currentInputSelected) then
	            vol = tonumber(tb["masterVolume"])
                mute = tonumber(tb["mute"])
	            g_currentVolume = vol
                if (mute == 1) then 
                    g_currentMuted = true
                else
                    g_currentMuted = false
                end
                printd("Sending VOLUME_LEVEL_CHANGED with level of " .. g_currentVolume .. " for RECEIVED_FROM_NETWORK.get_mastervolume")
	            SendNotify("VOLUME_LEVEL_CHANGED", {LEVEL = g_currentVolume}, NOTIFY_OUTPUT_INDEX)
                C4:UpdateProperty("Master Volume", vol)
                SendNotify("MUTE_CHANGED", {MUTE = g_currentMuted}, NOTIFY_OUTPUT_INDEX)
            end
        end
    end
end


function RECEIVED_FROM_NETWORK.processError(jrep)
    local ec = tostring(jrep["status"])
	if (Log._logLevel > 3) then 
        local ec = tostring(jrep["status"])
        print("EC = " .. ec)
        local estring = errorCode[ec]
        if (estring) then 
           print(jrep["type"] .. " -> " .. estring)
        else 
           print("no error code")
        end
	end
    if (ec == "200") then
        g_okCount = g_okCount + 1
    else
       g_errorCount = g_errorCount + 1
    end
    if (ec == "402") then
        g_needLoginRetry = true;
    end
end

function starts_with(str, start)
    return str:sub(1, #start) == start
end

function ReceivedFromNetwork(idBinding, nPort, strData)
	if (strData == nil) then 
		print(string.format("%.2f", socket.gettime() - g_startTime) .. ":<----- received nil packet ")
		return
	end
	if (Log._logLevel > 3) then 
		print(string.format("%.2f", socket.gettime() - g_startTime) .. ":<----- received data from device ")
		hexdump(strData)
	end
    g_unparsedJsonBuffer = g_unparsedJsonBuffer .. strData
    -- count the { - don't parse until we have same number of }
    local _, countStarts = string.gsub(g_unparsedJsonBuffer, "{", "")
    local countEnds = 0
    local endOfBuf = 0
    while true do
      endOfBuf = string.find(g_unparsedJsonBuffer, "}", endOfBuf + 1) 
      if endOfBuf == nil then break end
      countEnds = countEnds + 1
      if (countEnds == countStarts) then break end
    end
    --print("cs = " .. countStarts .. ", ce = " .. countEnds .. ", eob = " .. endOfBuf) 
    if (countStarts == countEnds and endOfBuf ~= nil and endOfBuf > 0) then
    	local jrep = JSON:decode(g_unparsedJsonBuffer)
        --trim the global data to the end of the pasrsing - we may have the start of a new packet in the buffer
        -- do this after the parse but before we handle stuff incase we throw and error
        g_unparsedJsonBuffer = g_unparsedJsonBuffer:sub(endOfBuf + 1)
        if (jrep) then
	        RECEIVED_FROM_NETWORK.processError(jrep)
	        local trimmedResponse = string.gsub(jrep["type"], " ", "")
	        if (RECEIVED_FROM_NETWORK[trimmedResponse] ~= nil and type(RECEIVED_FROM_NETWORK[trimmedResponse]) == "function") then
		        RECEIVED_FROM_NETWORK[trimmedResponse](jrep)
	        end
            if (starts_with(trimmedResponse, "unsol_")) then
                --string off the unsol_ and look for a matching RECEIVED_FROM_NETWORK call??
                printd("Received unsol_ message of " .. trimmedResponse)
            else
                -- let the protocol know we received a response and he can send next packet from queue
                epi_protocol.responseReceived()
            end
        end
    end
end

-------------------------------------------------------------------------------------------
--							SOME HELPER FUNCTIONS
-------------------------------------------------------------------------------------------
function GetProxyInputIndex(inputIndex)
   -- OneBasedOutputIndexToNotifyOuputId - we use a 1 based index, but we need a 4000 based index for the proxy
   return (inputIndex + 2999)
end

function GetProxyOutputIndex(outputIndex)
   -- OneBasedOutputIndexToNotifyOuputId - we use a 1 based index, but we need a 7000 based index for the proxy as these are only room volume endpoints
   --return (outputIndex + 6999)
   return NOTIFY_OUTPUT_INDEX
end

function roundHalf(number)
	floor = math.floor(number)
	if floor + .25 > number then
		return floor
	end
	if floor + .75 > number then
		return floor+.5
	end
	return math.ceil(number)
end

-- given a 0-100 based value, convert it to a -12-+12 value in .5 steps with rounding.  0 would be -12, 50 would be 0, etc.
function Convert100BasedValueToNeg12Pos12Based(hunVal)
   if (hunVal > 100) then return 12 end
   if (hunVal < 0) then return -12 end
   -- we has 49 discrete values (12x4 + 0).  
   local newVal = (hunVal * 48) / 200
   newVal = roundHalf(newVal)
   newVal = newVal - 12;
   return newVal
end

-- given a -12 to 12 based value, convert it to a 0-100 with rounding.  -12 would be 0, 0 would be 50, etc.
function ConvertNeg12Pos12BasedValueTo100Based(twelveVal)
   if (twelveVal > 12) then return 100 end
   if (twelveVal < -12) then return 0 end
   local newVal = ((twelveVal + 12) * 200) / 48
   newVal = roundHalf(newVal)
   return newVal
end

function GetInputFromParams(tParams)
   local zeroBased = tParams["INPUT"]
   if (zeroBased == nil) then zeroBased = tParams["input"] end
   zeroBased = tonumber(zeroBased)
   zeroBased = zeroBased % 100
   return zeroBased + 1
end

function GetOutputFromParams(tParams)
   -- since for today, only one output, return 1.  BUT IF MORE THAN ONE OUTPUT fix below!
   return 1
--   if (tParams == nil) then return 1 end
--   local zeroBased = tParams["OUTPUT"]
--   if (zeroBased == nil) then zeroBased = tParams["output"] end
--   if (zeroBased == nil) then zeroBased = tParams["BINDING_ID"] end
--   if (zeroBased == nil) then return 1 end
--   zeroBased = tonumber(zeroBased)
--   zeroBased = zeroBased % 100
--   return zeroBased + 1
end

function GetEqBandFromParams(tParams)
   local zeroBased = tonumber(tParams["EQ_INDEX"])
   return zeroBased + 1
end

function GetBoolFromParam(stringVal)
   if ((stringVal == nil) or (stringVal == false)) then
      return false
   end
   if ((stringVal == "false") or (stringVal == "False") or (stringVal == "FALSE")) then
      return false
   end
   return true
end

function SendOutputToInputDelayed(timer, skip)
	for key, value in pairs(g_delayedInputCmdsTimers) do
	    if (value == timer) then
		    epi_protocol.setOutputToInput(key, g_currentInputSelected)
		end
	end
end

function SetOutputToInputDelayed(outputIndex, inputIndex)
   if (g_delayedInputCmdsTimers[outputIndex] ~= nil) then g_delayedInputCmdsTimers[outputIndex]:Cancel() end
   g_delayedInputCmdsTimers[outputIndex] = C4:SetTimer(250, SendOutputToInputDelayed, false)
end

-------------------------------------------------------------------------------------------
--							ACTIONS
-------------------------------------------------------------------------------------------
function LUA_ACTION.ACTION_diagData()
	if (g_epiData.outputCount == 8) then
		diag_FilePrefix = C4:GetC4zDir() .. "triad_ams8/www/firmware/"
		diag_HttpPrefix = "/driver/triad_ams8/firmware/"
	end
	if (g_epiData.outputCount == 24) then	 
		diag_FilePrefix = C4:GetC4zDir() .. "triad_ams24/www/firmware/"
		diag_HttpPrefix = "/driver/triad_ams24/firmware/"
	end
	g_diagHttpEntry = "http://" .. C4:GetControllerNetworkAddress() .. diag_HttpPrefix .. "diagResults.txt"
	local diagFileEntry = "diagResults.txt"
	print("using dir of " .. diag_FilePrefix)
	C4:FileSetDir(diag_FilePrefix)
	local bexist = C4:FileExists(diagFileEntry)
	if (bexist) then
	   C4:FileDelete(diagFileEntry)
	end
	if ((g_diagFileHandle) and (g_diagFileHandle > 0)) then
	   print("Closing opened file handle for new run of diagnostics")
       C4:FileClose(g_diagFileHandle)
	   g_diagFileHandle = nil
	end
    g_diagFileHandle = C4:FileOpen(diagFileEntry)
	if (g_diagFileHandle == -1) then
		print("ERROR - unable to open file " .. diagFileEntry .. " for logging diagnostic data")
		return
	end
	print("starting to gather diagnostic data...")
	C4:FileWriteString(g_diagFileHandle, "-----------------------------  Device Data -----------------------------\n")
	epi_protocol.sendGetDiagnosticCommands()
end

function DiagnosticLogDriverData (t, indent, done)
  done = done or {}
  indent = indent or ''
  local nextIndent -- Storage for next indentation value
  for key, value in pairs (t) do
    if type (value) == "table" and not done [value] then
      nextIndent = nextIndent or
          (indent .. string.rep(' ',string.len(tostring (key))+2))
          -- Shortcut conditional allocation
      done [value] = true
      C4:FileWriteString(g_diagFileHandle, indent .. "[" .. tostring (key) .. "] => Table {\n")
      C4:FileWriteString(g_diagFileHandle, nextIndent .. "{\n")
      DiagnosticLogDriverData (value, nextIndent .. string.rep(' ',2), done)
      C4:FileWriteString(g_diagFileHandle, nextIndent .. "}\n");
    else
      C4:FileWriteString(g_diagFileHandle, indent .. "[" .. tostring (key) .. "] => " .. tostring (value).."\n")
    end
  end
end

function DiagnosticsLogData(result)
    result = string.gsub(result, "\n", "")
	if (string.len(result) > 0) then
		result = result .. "\n"
		C4:FileWriteString(g_diagFileHandle, result)
		if (string.find(result,"Get MAC Add") ~= nil) then
			-- we are done with device - put in the driver state
			C4:FileWriteString(g_diagFileHandle, "-----------------------------  Driver Data -----------------------------\n")
			DiagnosticLogDriverData(g_epiData)
			C4:FileClose(g_diagFileHandle)
			g_diagFileHandle = nil
			print("Diagnostic data gathering completed - download results at: " .. g_diagHttpEntry)
		end
	end
end


function LUA_ACTION.ACTION_logDuration()
  print("Extending log duration for 24 hours - changing log levels will reset time to 5 minutes")
  if (g_printTimerId ~= nil) then g_printTimerId:Cancel() end
  g_printTimerId = C4:SetTimer(1000 * 60 * 60 * 24, LogTimerHandler)
end

-------------------------------------------------------------------------------------------
--							EXECUTE COMMANDS
-------------------------------------------------------------------------------------------
function LogTimerHandler(timer, skips)
   print("print timer expired - resetting logging")
   timer:Cancel()
   Log:OutputPrint(false)
   Log:OutputC4Log(false)
   Log:SetLogLevel(2)
   C4:UpdateProperty("Debug Mode", "Off")
   C4:UpdateProperty("Debug Level", "2 - Warning")
end

function EX_CMD.REQUEST_ALL_SETTINGS(tParams)
   g_shouldSuppress = false
   SendGuiAllData()
end

function EX_CMD.SET_NETWORK_SETTINGS(tParams)
	g_epiData.DHCP = tParams["DHCP"]
	if (g_epiData.DHCP == "True") then
		g_epiData.DHCP = true
	else
		g_epiData.DHCP = false
	end
	g_epiData.IPADDRESS = tParams["IPADDRESS"]
	g_epiData.GATEWAYADDRESS = tParams["GATEWAYADDRESS"]
	g_epiData.MASKADDRESS = tParams["MASKADDRESS"]
    C4:UpdateProperty("Web Config", "http://" .. g_epiData.IPADDRESS)
	if (g_epiData.DHCP) then 
		epi_protocol.setIpAddressDHCP()
	else
		epi_protocol.setIpAddressStatic(g_epiData.IPADDRESS, g_epiData.MASKADDRESS, g_epiData.GATEWAYADDRESS)
	end
	NotifyNetworkState()
end

function VersionAtLeast(majorCheck, minorCheck, revCheck)
    local tVers = C4:GetVersionInfo()
    local directorVersion = tVers["version"]
    local major, minor, rev, build = string.match(directorVersion, "(%d+)%.(%d+)%.(%d+)%.(%d+)")
    if (tonumber(major) > tonumber(majorCheck)) then return true end
    if (tonumber(major) == tonumber(majorCheck) and tonumber(minor) > tonumber(minorCheck)) then return true end
    if (tonumber(major) == tonumber(majorCheck) 
         and tonumber(minor)  == tonumber(minorCheck) 
         and tonumber(rev) >= tonumber(revCheck)) then
            return true
    end
    return false
end

function EX_CMD.SET_VOLUME_LEVEL(tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentVolume = tonumber(tParams["LEVEL"])
   Volume(outputIndex, g_currentVolume)
end

function EX_CMD.SET_MUTE_STATE(tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentMuted = MuteStateToValue(GetBoolFromParam(tParams["Muted"]))
   Mute(outputIndex, g_currentMuted)
end

function EX_CMD.SET_SUPRESS_NOTIFIES(tParams)
      g_shouldSuppress = tParams["ENABLED"]
end

function EX_CMD.SET_DISABLE_AUDIO_SENSE(tParams)
      local isTrue = tostring(tParams["DISABLED"])
	  if ((isTrue == "True") or (isTrue == "true")) then
		g_epiData.disableAudioSense = true
	  else
		g_epiData.disableAudioSense = false
	  end
      epi_protocol.disableAudioSense(g_epiData.disableAudioSense)
end

function EX_CMD.SET_AUDIO_SENSE_OFF_DELAY(tParams)
   g_epiData.audioSenseOffDelay = tonumber(tParams["LEVEL"])
   epi_protocol.setAudioSenseOffDelay(g_epiData.audioSenseOffDelay)
end

function PRX_CMD.SET_AUDIO_SENSE_OFF_DELAY (bindingId, tParams)
   g_epiData.audioSenseOffDelay = tonumber(tParams["LEVEL"])
   epi_protocol.setAudioSenseOffDelay(g_epiData.audioSenseOffDelay)
end

function EX_CMD.SET_FACTORY_DEFAULTS(tParams)
	printd("SET_FACTORY_DEFAULTS setting up new data")
	-- setup data object
	outputCount = tonumber(C4:GetCapability("audio_provider_count"))
	inputCount = tonumber(C4:GetCapability("audio_consumer_count"))
	g_epiData = epi_data.Init(outputCount, inputCount)
	epi_protocol.setFactoryDefaults()
	epi_data.SyncDeviceToData(g_epiData)
	-- notify proxy of default state
	SendProxyAllNotifyData()
	epi_protocol.getFirmwareVersion()
end

function EX_CMD.GET_FIRMWARE_VERSION(tParams)
	epi_protocol.getFirmwareVersion()
end

-------------------------------------------------------------------------------------------
--							EXECUTE_CMD and PROXY NOTIFICATIONS
-------------------------------------------------------------------------------------------
function SendGuiAllData()
   if (g_shouldSuppress) then return end
   if (g_epiData == nil) then
      print("calling OnDriverInit() from EX_CMD.REQUEST_ALL_SETTINGS")
      OnDriverInit()
   end
end

function NotifyStartAudioSense(inputIndex)
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, "START_AUDIO_SENSE", {INPUT = GetProxyInputIndex(inputIndex)}, "NOTIFY")
end

function NotifyStopAudioSense(inputIndex)
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, "STOP_AUDIO_SENSE", {INPUT = GetProxyInputIndex(inputIndex)}, "NOTIFY")
end

function NotifyNetworkState()
   if (g_shouldSuppress) then return end
   local networkState = {}
   networkState._COMMAND = "I:NETWORK_STATE"
   networkState.DHCP = g_epiData.DHCP
   networkState.IPADDRESS = g_epiData.IPADDRESS
   networkState.GATEWAYADDRESS = g_epiData.GATEWAYADDRESS
   networkState.MASKADDRESS = g_epiData.MASKADDRESS
end

function SendNotify(notifyCmd, tParams, outputIndex)
	if (Log._logLevel > 3) then 
		print("SendNotify: notifyCmd=" .. notifyCmd .. " outIndex = " .. outputIndex)
		PrintTable(tParams, "tParams->")
	end
	tParams.OUTPUT = tostring(GetProxyOutputIndex(outputIndex))
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, notifyCmd, tParams, "NOTIFY")
end

function SendNotifyInput(notifyCmd, tParams, inputIndex)
	if (Log._logLevel > 3) then 
		print("SendNotify: notifyCmd=" .. notifyCmd .. " inputIndex = " .. inputIndex)
		PrintTable(tParams, "tParams->")
	end
	tParams.INPUT = tostring(GetProxyInputIndex(inputIndex))
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, notifyCmd, tParams, "NOTIFY")
end

-------------------------------------------------------------------------------------------
--							PROXY COMMANDS - Handlers
-------------------------------------------------------------------------------------------
----------------------- Volume -----------------------
g_volTimers= {}

function NotifyVolumeLevel(outputIndex, suppressNotify)
   if (suppressNotify == nil) then suppressNotify = false end
   if (suppressNotify == false) then
      SendNotify("VOLUME_LEVEL_CHANGED", {LEVEL = g_currentVolume}, outputIndex)
   end
end

function StartVolumeTimer(outputIndex, goingUp)
   if (g_volTimers[outputIndex] ~= nil) then g_volTimers[outputIndex].timerId:Cancel() end
   g_volTimers[outputIndex] = {}
   g_volTimers[outputIndex].goingUp = goingUp
   g_volTimers[outputIndex].timerId = C4:SetTimer(300, VolumeTimerHandler, true)
end

function VolumeTimerHandler(timer, skips)
   for outputIndex, timerData in pairs(g_volTimers) do 
      if (timerData.timerId == timer) then
	   if (timerData.goingUp) then
               g_currentVolume = g_currentVolume + 1
        else
               g_currentVolume = g_currentVolume  - 1
        end
        CheckMaxVolume(outputIndex)
        if (Volume(outputIndex, g_currentVolume)) then timer:Cancel() end
       end
   end
end

function PRX_CMD.START_VOL_UP(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentVolume = g_currentVolume + 1
   CheckMaxVolume(outputIndex)
   Volume(outputIndex, g_currentVolume)
   StartVolumeTimer(outputIndex, true)
end

function PRX_CMD.STOP_VOL_UP(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   if (g_volTimers[outputIndex] ~= nil) then g_volTimers[outputIndex].timerId:Cancel() end
   g_volTimers[outputIndex] = nil
end

function PRX_CMD.START_VOL_DOWN(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentVolume = g_currentVolume- 1
   Volume(outputIndex, g_currentVolume)
   StartVolumeTimer(outputIndex, false)
end

function PRX_CMD.STOP_VOL_DOWN(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   if (g_volTimers[outputIndex] ~= nil) then g_volTimers[outputIndex].timerId:Cancel() end
   g_volTimers[outputIndex] = nil
end

function PRX_CMD.SET_VOLUME_LEVEL(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentVolume = tonumber(tParams["LEVEL"])
   CheckMaxVolume(outputIndex)
   Volume(outputIndex, g_currentVolume)
end

function PRX_CMD.PULSE_VOL_UP(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentVolume = g_currentVolume + 1
   CheckMaxVolume(outputIndex)
   Volume(outputIndex, g_currentVolume)
end

function PRX_CMD.PULSE_VOL_DOWN(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   if (g_currentVolume > 0) then
	   g_currentVolume =g_currentVolume - 1
	   Volume(outputIndex, g_currentVolume)
	end
end

function PRX_CMD.BINDING_CHANGE_ACTION (bindingId, tParams)
	-- we don't want to rest volumes on project load - so only apply this logic after late Init
	local state = tParams["IS_BOUND"]
	local outputIndex = GetOutputFromParams(tParams)
	local outputBinding = tonumber(tParams["BINDING_ID"])
	Log:Debug("BINDING_CHANGE_ACTION for  " .. outputBinding .. " to state of " .. state)
	-- we keep a map of the active volume bound outputs and don't set the volume for audio out if volume is bound
	if ((outputBinding > 6999) and (outputBinding < 8000)) then
		g_volumeOutputIsBoundMap[outputBinding] = state
	end

	if (g_lateInited == true) then
		if ((outputBinding > 3999) and (outputBinding < 5000)) then
			if (state == "True") then
			    if (g_volumeOutputIsBoundMap[outputBinding + 3000] ~= "True") then
					Log:Info("Setting output " .. outputIndex .. " to default volume of " .. DEFAULT_BINDING_VOLUME_NON_ROOM .. " due to non-room binding change")
					g_currentVolume = DEFAULT_BINDING_VOLUME_NON_ROOM
					CheckMaxVolume(outputIndex)
					Volume(outputIndex, g_currentVolume)
				else
					Log:Info("Not setting output " .. outputIndex .. " to default volume due to room volume being bound")
				end
			end
		elseif ((outputBinding > 6999) and (outputBinding < 8000)) then
			if (state == "True") then
				Log:Info("Setting output " .. outputIndex .. " to default volume of " .. DEFAULT_BINDING_VOLUME_ROOM .. " due to room volume binding change")
				g_currentVolume = DEFAULT_BINDING_VOLUME_ROOM
				CheckMaxVolume(outputIndex)
				Volume(outputIndex, g_currentVolume)
			end
		end
	end
end

----------------------- MUTE -----------------------
function PRX_CMD.MUTE_ON(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentMuted = MuteStateToValue("Muted")
   Mute(outputIndex, g_currentMuted)
end

function PRX_CMD.MUTE_OFF(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_currentMuted = MuteStateToValue("Unmuted")
   Mute(outputIndex, g_currentMuted)
end

function PRX_CMD.MUTE_TOGGLE(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   if (g_currentMuted == "Muted") then
        g_currentMuted = MuteStateToValue("Unmuted")
   else
        g_currentMuted = MuteStateToValue("Muted")
   end
   Mute(outputIndex, g_currentMuted)
   C4:UpdateProperty("Muted", g_currentMuted)
end

----------------------- LOUDNESS  -----------------------
function PRX_CMD.LOUDNESS_ON(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_epiData.outputs[outputIndex].outputLoudnessEnabled = true
   OutputLoudness(outputIndex, g_epiData.outputs[outputIndex].outputLoudnessEnabled)
end

function PRX_CMD.LOUDNESS_OFF(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_epiData.outputs[outputIndex].outputLoudnessEnabled = false
   OutputLoudness(outputIndex, g_epiData.outputs[outputIndex].outputLoudnessEnabled)
end

function PRX_CMD.LOUDNESS_TOGGLE (bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_epiData.outputs[outputIndex].outputLoudnessEnabled = not g_epiData.outputs[outputIndex].outputLoudnessEnabled
   OutputLoudness(outputIndex, g_epiData.outputs[outputIndex].outputLoudnessEnabled)
end


----------------------- Surround Mode  -----------------------
function PRX_CMD.SET_SURROUND_MODE(bindingId, tParams)
   if (g_currentInputSelected == 0) then return end
   local smode = tonumber(tParams["SURROUND_MODE"]) - 1
   g_epiData.inputs[g_currentInputSelected].soundMode = smode
   epi_protocol.setSurroundMode(g_currentInputSelected, g_epiData.inputs[g_currentInputSelected].soundMode)
   SendNotifyInput("SURROUND_MODE_CHANGED",  {SURROUND_MODE = tostring(smode)}, NOTIFY_OUTPUT_INDEX)
end

----------------------- INPUTS and OUTPUTS -----------------------
function EX_CMD.SET_INPUT_GAIN_LEVEL(tParams)
   local inputIndex = GetInputFromParams(tParams)
   g_epiData.inputs[inputIndex] = tonumber(tParams["LEVEL"])
   epi_protocol.setInputGain(inputIndex, g_epiData.inputs[inputIndex])
   SendNotifyInput("INPUT_GAIN_CHANGED",  {LEVEL = tostring(g_epiData.inputs[inputIndex])}, inputIndex)
end

function PRX_CMD.OFF()
    Log:Debug("Received OFF from proxy")
    --if (g_useAutoArcSelect) then
     --  InputToOutput(AUTO_ARC_INPUT_SELECTID, NOTIFY_OUTPUT_INDEX)
    --end
    epi_protocol.setPowerOff()
end

function PRX_CMD.ON()
  -- theory here is that we should get the on, then the connect_output command for non-arc use
  -- the g_currentInputSelected will be set to AUTO_ARC_INPUT_SELECTID here, and the
    Log:Debug("Received ON from proxy")
    epi_protocol.setPowerOn()
    if (g_useAutoArcSelect) then
        InputToOutput(AUTO_ARC_INPUT_SELECTID, NOTIFY_OUTPUT_INDEX)
    end
end

function PRX_CMD.CONNECT_OUTPUT(bindingId, tParams)
-- may need this for auto arc selection to work - tdb
end

function PRX_CMD.SET_INPUT(bindingId, tParams)
   local inputIndex = GetInputFromParams(tParams)
   local outputIndex = GetOutputFromParams(tParams)
   InputToOutput(inputIndex, outputIndex)
end

function PRX_CMD.DISCONNECT_OUTPUT(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   DisconnectOutput(outputIndex)
end

function PRX_CMD.SET_INPUTGAIN_LEVEL(bindingId, tParams)
   local inputIndex = GetInputFromParams(tParams)
   g_epiData.inputs[inputIndex] = tonumber(tParams["LEVEL"])
   epi_protocol.setInputGain(inputIndex, g_epiData.inputs[inputIndex])
   SendNotifyInput("INPUT_GAIN_CHANGED",  {LEVEL = tostring(g_epiData.inputs[inputIndex])}, inputIndex)
end


function PRX_CMD.GET_AUDIO_INPUTS(bindingId, tParams)
   -- my belief is that only Composer cares about this to sync up the proxy avswitch code for ctrl-ctrl
   -- since we give all input/output status up differently, not needed? 
   print("GET_AUDIO_INPUTS proxy command called - do I need to handle this?")
end

function PRX_CMD.QUERY_OUTPUT_STATUS(bindingId, tParams)
   -- my belief is that only Composer cares about this to sync up the proxy avswitch code for ctrl-ctrl
   -- since we give all input/output status up differently, not needed? 
   print("QUERY_OUTPUT_STATUS proxy command called - do I need to handle this?")
end

function PRX_CMD.QUERY_ALL_OUTPUT_STATUS(bindingId, tParams)
   -- my belief is that only Composer cares about this to sync up the proxy avswitch code for ctrl-ctrl
   -- since we give all input/output status up differently, not needed? 
   print("QUERY_ALL_OUTPUT_STATUS proxy command called - do I need to handle this?")
end

function PRX_CMD.SET_ROOM_BINDING_NAME(bindingId, tParams)
   local outputIndex = GetOutputFromParams(tParams)
   g_roomBindingNames[outputIndex] = tParams["PartnerDevice"]
end

function CheckMaxVolume(outputIndex)
   if (g_currentVolume > MAX_VOL_VALUE) then
      g_currentVolume = MAX_VOL_VALUE
      printd("Sending VOLUME_LEVEL_CHANGED with level of " .. g_currentVolume .. " for CheckMaxVolume")
      SendNotify("VOLUME_LEVEL_CHANGED", {LEVEL = g_currentVolume}, outputIndex)
   end
end

function Volume(outputIndex, volume, suppressNotify)
    if (supressNotify == nil) then supressNotify = false end
    g_currentVolume = volume
	local overLimit = false
	CheckMaxVolume(outputIndex)
	epi_protocol.setVolume(g_currentInputSelected, g_currentVolume)
	if (g_currentMuted) then
        Mute(outputIndex, "Unmuted")
	end
	overLimit = NotifyVolumeLevel(outputIndex, suppressNotify)
    C4:UpdateProperty("Master Volume", tostring(g_currentVolume))
	return overLimit
end

function Mute(outputIndex, isMuted)
    local notifyVal = "False"
    if (isMuted == "Muted" or isMuted == "true" or isMuted == true) then
       isMuted = "1"
       notifyVal = "True"
    else
       isMuted = "0"
    end
	epi_protocol.setOutputMuteState(g_currentInputSelected, isMuted)
	--NotifyVolumeLevel(outputIndex)
	SendNotify("MUTE_CHANGED", {MUTE = notifyVal}, outputIndex)
end

function MaxVolume(outputIndex, maxVolume)
	epi_protocol.setMaxVolume(outputIndex, maxVolume)
    CheckMaxVolume(outputIndex)
end
	
function InputToOutput(inputIndex, outputIndex)
    g_currentInputSelected = inputIndex
    print("Received InputToOutput with input of " .. tonumber(inputIndex) .. " and output of " .. tonumber(outputIndex) .. " from Proxy")
    epi_protocol.setPowerOn()
    SetOutputToInputDelayed(outputIndex, inputIndex)
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, 'INPUT_OUTPUT_CHANGED', {INPUT = (GetProxyInputIndex(g_currentInputSelected)), OUTPUT = GetProxyOutputIndex(outputIndex), AUDIO=true, VIDEO=false}, "NOTIFY")
	Volume(outputIndex, g_currentVolume, true)
    C4:SetVariable(DV_ACTIVE_ZONE_NAME, 1)
end

function DisconnectOutput(outputIndex)
    g_currentInputSelected = AUTO_ARC_INPUT_SELECTID
    if (g_delayedInputCmdsTimers[outputIndex] ~= nil) then g_delayedInputCmdsTimers[outputIndex]:Cancel() end
	C4:SendToProxy(DEFAULT_PROXY_BINDING_ID, 'INPUT_OUTPUT_CHANGED', {INPUT = -1, OUTPUT = GetProxyOutputIndex(outputIndex), AUDIO=true, VIDEO=false}, "NOTIFY")  
    epi_protocol.disconnectOutput(outputIndex)
    local isActive = 0
    C4:SetVariable(DV_ACTIVE_ZONE_NAME, isActive)
end

-- Some Helper Functions	

-- this function is to handle the bad data from a bug in composer where locale data with a comma was 
-- send for decimal data
function FixCommanDecimalValues(tValue)
  for k,v in pairs(tValue) do
    --print(sIndent .. tostring(k) .. ":  " .. tostring(v))
    if (type(v) == "table") then
      FixCommanDecimalValues(v)
    end
	if ((k == "gain") or (k == "balance") or (k == "treble") or (k == "bass")) then
	   if (string.match(v, ",")) then
	      vs = tostring(v)
	      print("Correcting invalid data " .. vs .. " in setting " .. k)
		  print("before: " .. tValue[k])
		  tValue[k] = vs:gsub(",", ".")
		  print("after: " .. tValue[k])
	   end
	end
  end
end

function tablelength(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end
-- declare local variables
--// exportstring( string )
--// returns a "Lua" portable version of the string
local function exportstring( s )
    return string.format("%q", s)
end


function myFileWrite(fileHandle, str)
   return(C4:FileWrite(fileHandle, string.len(str), str))
end

--// The Save Function
function table.c4save(tbl, filename)
	local charS,charE = "   ","\n"
	if (C4:FileExists(filename)) then C4:FileDelete(filename) end
	local file = C4:FileOpen(filename)
	if (file == -1) then
		print("unable to open file " .. filename .. " to save table")
		return
	end
	print("opened file " .. filename .. " to save table")

	-- initiate variables for save procedure
	local tables,lookup = { tbl },{ [tbl] = 1 }
	myFileWrite(file, "return {"..charE)

	for idx,t in ipairs( tables ) do
		myFileWrite(file, "-- Table: {"..idx.."}"..charE )
		myFileWrite(file, "{"..charE )
		local thandled = {}

		for i,v in ipairs( t ) do
			thandled[i] = true
			local stype = type( v )
			-- only handle value
			if stype == "table" then
			if not lookup[v] then
			table.insert( tables, v )
			lookup[v] = #tables
		end
		myFileWrite(file, charS.."{"..lookup[v].."},"..charE )
		elseif stype == "string" then
			myFileWrite(file, charS..exportstring( v )..","..charE )
		elseif stype == "number" then
			myFileWrite(file, charS..tostring( v )..","..charE )
		end
	end

	for i,v in pairs( t ) do
		-- escape handled values
		if (not thandled[i]) then
			local str = ""
			local stype = type( i )
			-- handle index
			if stype == "table" then
				if not lookup[i] then
					table.insert( tables,i )
					lookup[i] = #tables
				end
				str = charS.."[{"..lookup[i].."}]="
				elseif stype == "string" then
					str = charS.."["..exportstring( i ).."]="
				elseif stype == "number" then
					str = charS.."["..tostring( i ).."]="
				end
				if str ~= "" then
					stype = type( v )
					-- handle value
					if stype == "table" then
						if not lookup[v] then
						table.insert( tables,v )
						lookup[v] = #tables
						end
					myFileWrite(file, str.."{"..lookup[v].."},"..charE )
					elseif stype == "string" then
						myFileWrite(file, str..exportstring( v )..","..charE )
					elseif stype == "number" then
						myFileWrite(file, str..tostring( v )..","..charE )
					end
				end
			end
		end
		myFileWrite(file, "},"..charE )
	end
	myFileWrite(file, "}" )
	C4:FileClose(file)
	print("closed file " .. filename .. " to save table")
end


function LinkedList_validate(list)
	local entries = {}
	local item = list._start
	local prevItem, lastItem
	while item do
		prevItem = item._prev
		if item._prev ~= prevItem then
			print("corrupted: discovered incorrect prev pointer")
			return false
		end
		prevItem = item._prev
		for i = 1, #entries do
			if entries[i] == item then
				print("corrupted: found next pointer to item already iterated")
				return false
			end
		end
		table.insert(entries, item)
		lastItem = item
		item = item._next
	end
	if list._end ~= lastItem then
		print("corrupted: end pointer not pointing to last item")
		return false
	end
	if #entries ~= list._length then
		print("corrupted: discovered " .. #entries .. " entries, but should be " .. list._length)
		return false
	end
		print("not corrupted")
	return true
end

--LinkedList_validate(RhapsodyClient._cache._list)
--table.c4save(RhapsodyClient._cache, "bb.xml")

	
function PrintTableSizes(tValue, sIndent)
  for k,v in pairs(tValue) do
    --print(sIndent .. tostring(k) .. ":  " .. tostring(v))
    if (type(v) == "table") then
      print(sIndent .. tostring(k) .. ":  " .. tostring(v) .. " size->" .. tablelength(v))
      PrintTable(v, sIndent .. " ")
    end
  end
end

function printd(str)
    if (Log._logLevel > 3) then
	    print(str)
    end
end

------------------------  Global init -----------------------------

C4:AllowExecute(true)
g_errorCount = 0
g_okCount = 0

do -- LOCALE FIXING FOR tostring AND tonumber

	if (not tostring_native) then
		tostring_native = tostring
	end
	if (not tonumber_native) then
		tonumber_native = tonumber
	end

	LOCALE_USES_COMMA_DECIMAL_SEPARATORS = (tonumber ('0.5') == nil)

	function tostring_return_comma (v)
		local ret = tostring_native (v)
		if (type (v) == 'number') then
			ret = string.gsub (ret, '%.', '%,')
		end
		return (ret)
	end

	function tostring_return_period (v)
		local ret = tostring_native (v)
		if (type (v) == 'number') then
			ret = string.gsub (ret, '%,', '%.')
		end
		return (ret)
	end

	function tonumber_expect_comma (e, base)
		local ret = tonumber_native (e, base)
		if (ret == nil) then
			if (type (e) == 'string') then
				e = string.gsub (e, '%.', '%,')
				ret = tonumber_native (e, base)
			end
		end
		return (ret)
	end

	function tonumber_expect_period (e, base)
		local ret = tonumber_native (e, base)
		if (ret == nil) then
			if (type (e) == 'string') then
				e = string.gsub (e, '%,', '%.')
				ret = tonumber_native (e, base)
			end
		end
		return (ret)
	end

	if (LOCALE_USES_COMMA_DECIMAL_SEPARATORS) then
		tonumber = tonumber_expect_comma
	end
end

-- either assign the exisiting persit data to the g_epiData or create a new empty one
if (PersistData ~= nil) then
    print("PersistData found - restoring state")
    g_epiData = PersistData
    OnDriverInit()
    for k, _ in pairs (Properties) do
		OnPropertyChanged (k)
	end

else
    print("No PersistData found")
    OnDriverInit()
end

C4:AllowExecute(true)

gIsDevelopmentVersionOfDriver = true
