--------------------------------------------------------------------------------------------
--	AudioQueueItem.lua
--
--  Copyright 2014 Control4 Corporation. All rights reserved.
--	
--	Represents a playable item in the audio queue.
-- 
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"
require "law.Xml"
require "law.StringBuilder"
require "law.StringUtils"

AudioQueueItem = inheritsFrom(nil)

AudioQueueItem.nonce = 0

--[[
	Constructor
--]]
function AudioQueueItem:init()
	self._title = ""
	self._album = ""
	self._artist = ""
	self._genre = ""
	self._trackId = ""
	self._albumId = ""
	self._artistId = ""
	self._playbackSeconds = 0
	
	self._id = AudioQueueItem.nonce

	-- Make sure we only use the nonce one time
	AudioQueueItem.nonce = AudioQueueItem.nonce + 1

end

--[[
	Converts the audio item to xml. 
	@param actionIds A list of space delimeted actions the list item should have.
--]]
function AudioQueueItem:toXml(actionIds)
	local sb = StringBuilder:new()

	sb:add("<item>")		
		
		sb:add("<title>")
		sb:add(C4:XmlEscapeString(self:getTitle()))
		sb:add("</title>")
		
		sb:add("<trackId>")
		sb:add(C4:XmlEscapeString(self:getTrackId()))
		sb:add("</trackId>")

		-- This is needed to change the now playing item.
		sb:add("<Id>")
		sb:add(C4:XmlEscapeString(self:getId()))
		sb:add("</Id>")
		
		sb:add("<album>")
		sb:add(C4:XmlEscapeString(self:getAlbum()))
		sb:add("</album>")

		sb:add("<albumId>")
		sb:add(C4:XmlEscapeString(self:getAlbumId()))
		sb:add("</albumId>")

		sb:add("<artist>")
		sb:add(C4:XmlEscapeString(self:getArtist()))
		sb:add("</artist>")

		local artist = self:getArtist()	
		local album = self:getAlbum()
		if (artist and artist ~= "") or (album and album ~= "") then
			sb:add("<info>")

			if artist and artist ~= "" then
				sb:add(C4:XmlEscapeString(self:getArtist()))
			end

			if (artist and artist ~= "") or (album and album ~= "") then
				sb:add(" - ")
			end

			if album and album ~= "" then
				sb:add(C4:XmlEscapeString(self:getAlbum()))
			end

			sb:add("</info>")
		end	

		sb:add("<artistId>")
		sb:add(C4:XmlEscapeString(self:getArtistId()))
		sb:add("</artistId>")

		sb:add("<url>")
		sb:add(C4:XmlEscapeString(self:getUrl()))
		sb:add("</url>")

		sb:add("<playbackTime>")
		sb:add(C4:XmlEscapeString(self:secondsToTime()))
		sb:add("</playbackTime>")

		local images = self:getImages()
		for i,image in ipairs(images or {}) do
			sb:add("<image" )
			sb:add( [[ width="]]  ) :add( image.width  ):add( [["]]  )
			sb:add( [[ height="]] ) :add( image.height ):add( [[">]] )
			sb:add(image.url)
			sb:add("</image>")
		end

		if actionIds then
			sb:add("<actionIds>")
			sb:add(actionIds)
			sb:add("</actionIds>")
		end

	sb:add("</item>")

	return tostring(sb)
end




--[[
	Gets a unique number that is assigned to this item in the constructor.
--]]
function AudioQueueItem:getId()
	return tostring(self._id)
end

function AudioQueueItem:setTitle(title)
	self._title = title
end

function AudioQueueItem:getTitle()
	return self._title
end

function AudioQueueItem:setTrackId(id)
	self._trackId = id
end

function AudioQueueItem:getTrackId()
	return self._trackId
end

function AudioQueueItem:setAlbum(name)
	self._album = name
end

function AudioQueueItem:getAlbum()
	return self._album
end

function AudioQueueItem:setAlbumId(id)
	self._albumId = id
end

function AudioQueueItem:getAlbumId()
	return self._albumId
end

function AudioQueueItem:setArtist(name)
	self._artist = name
end

function AudioQueueItem:getArtist()
	return self._artist
end

function AudioQueueItem:setArtistId(id)
	self._artistId = id
end

function AudioQueueItem:getArtistId()
	return self._artistId
end

function AudioQueueItem:setUrl(url)
	self._url = url
end

function AudioQueueItem:getUrl()
	return self._url
end

function AudioQueueItem:setPlaybackSeconds(timeStr)
	self._playbackSeconds = tonumber(timeStr)
end

function AudioQueueItem:getPlaybackSeconds()
	return self._playbackSeconds
end

function AudioQueueItem:secondsToTime()
	if self._playbackSeconds == nil then
		return ""
	end

	if self._playbackSeconds > 0 then
		return StringUtils.secondsToTime(self._playbackSeconds)
	else
		return ""
	end
end

function AudioQueueItem:setImages(arrImages)
	self._images = arrImages
end

function AudioQueueItem:getImages()
	return self._images
end

--[[
	Finds the image with the highest pixel count from the set of images

	@return The images table or nil if none are found
--]]
function AudioQueueItem:findLargestImageUrl()
	if not self._images or #self._images == 0 then
		return ""
	end

	local pixelCount = 0
	local largestImg = self._images[1]

	for i, image in ipairs(self._images or {}) do
		local imgPixelCount = image.width * image.height
		if imgPixelCount > pixelCount then
			pixelCount = imgPixelCount
			largestImg = image
		end
	end

	Log:Trace("AudioItem largest image:", largestImg.url)
	return largestImg.url
end