--------------------------------------------------------------------------------------------
--	AudioQueueManager.lua
--
--  Copyright 2014 Control4 Corporation. All rights reserved.
--	
--	Manages RoomAudioQueues.
-- 
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"
require "law.Xml"
require "law.RoomAudioQueue"
require "law.StreamAudioQueue"
require "law.ServiceLocator"
require "law.VariableManager"
require "law.EventDispatcher"

AudioQueueManager = inheritsFrom(EventDispatcher)

AudioQueueManager.EVENTS = {
	RoomSettingsChanged = "roomSettingsChanged", -- Sends the room id that changed.
}

AudioQueueManager.DigitalAudioDeviceId = 100002
AudioQueueManager.RoomMapVarId = 1009
AudioQueueManager.QueueSettingsVarId = 1003

--[[
	Constructor 
	@param proxyBindingId The id of the proxy to send commands to.
--]]
function AudioQueueManager:init(proxyBindingId)
	Log:Debug("AudioQueueManager:init()")
	-- Setup event dispatcher.
	self:super():init()

	self._proxyBindingId = proxyBindingId

	-- He holds the queues.
	self._queues = {}

	-- Shuffle and repeat settings by room. This info is created from digital audio queue info.
	-- Updated in updateQueueSettings().
	self._roomSettings = {}

	-- Holds a map of queue ids keys and an array of roomIds as values.
	self._queueRoomMap = {}

	-- Create delegates for digital audio variable callbacks
	self._roomMapCallback = function(roomMap) 
		self:updateRoomMap(roomMap)
	end

	self._queueSettingsCallback = function(queueSettings)
		self:updateQueueSettings(queueSettings)
	end

	self:_watchVariables()
end

function AudioQueueManager:_watchVariables()
	-- Watch room variables.  This might fail if called before OnDriverLateInit, but we'll do it again when we get that event.
	if ServiceLocator.VariableManager then
		ServiceLocator.VariableManager:registerVariable(AudioQueueManager.DigitalAudioDeviceId, AudioQueueManager.RoomMapVarId, self._roomMapCallback)
		ServiceLocator.VariableManager:registerVariable(AudioQueueManager.DigitalAudioDeviceId, AudioQueueManager.QueueSettingsVarId, self._queueSettingsCallback)
	else
		error("AudioQueueManager:init() requires ServiceLocator.VariableManager!")
	end
end

--[[
	Gets a queue for a room. 
	@param roomId The id of the room the queue is in
	@param createIfNoneExists Create the queue for the room if one doesn't exist
	@param queueType The type of the queue to create if createIfNoneExists is true or the type
		to convert the queue to if one was found. If nil the RoomAudioQueue type will be used
		when creating a new queue and no conversion will be done when a queue is found.

	@return queue Returns the queue or nil if none was found or created
--]]
function AudioQueueManager:getQueueForRoom(roomId, createIfNoneExists, queueType)

	for i, queue in ipairs(self._queues) do
		if queue:isInRoom(roomId) then
			if queueType then 
				return self:_convertQueueToType(queue, queueType)
			else
				return queue
			end
		end
	end

	if createIfNoneExists then
		if not queueType then
			queueType = RoomAudioQueue
		end

		local queue = queueType:new(roomId, self._proxyBindingId)
		self._queues[#self._queues + 1] = queue

		return queue
	end

	return nil
end

--[[
	Converts a queue to the new type if applicable. Queue must 
	be managed by this manager
--]]
function AudioQueueManager:_convertQueueToType(queue, newType)
	if not queue then
		error("AudioQueueManager:_convertQueueToType() queue is nil")
	end

	local queueIndex = 0

	for i,q in ipairs(self._queues) do
		if q == queue then
			queueIndex = i
			break
		end		
	end

	if queueIndex == 0 then
			error("AudioQueueManager:convertQueueToType, Cannot find queue")
	end

	if queue:isa(RoomAudioQueue) then
		if newType == StreamAudioQueue then
			local stream = StreamAudioQueue:new("", 0)
			stream:copy(queue)
			queue:destroy()
			self._queues[queueIndex] = stream
			-- Update the UIs
			stream:queueChanged()
			stream:dashboardChanged()
			return stream
		else
			return queue
		end
	else -- Stream audio type to queue
		if newType == RoomAudioQueue then
			local newQueue = RoomAudioQueue:new("", 0)
			newQueue:copy(queue)
			newQueue:addAudioItems(queue:getHistory(), RoomAudioQueue.AddFlags.Play)
			self._queues[queueIndex] = newQueue
			queue:destroy()
			-- Update the UIs
			newQueue:queueChanged()
			newQueue:dashboardChanged()
			return newQueue
		else
			return queue
		end
	end
end

--[[
	Gets the queue given its DigitalAudio id. 
--]]
function AudioQueueManager:getQueueById(queueId)
	for i, queue in ipairs(self._queues) do
		if queue:getQueueId() == queueId then
			return queue
		end
	end

	return nil
end

--[[
	Destroys all audio queues. 
	@param roomOff Tells the queue to turnoff all associated rooms
--]]
function AudioQueueManager:clearQueues(roomOff)
	-- Make a shallow copy of the array.
	-- We do this just in case the destroy has side effects that would change the queues array.
	local queues = tableConcat({}, self._queues)

	for i,queue in ipairs(queues) do
		queue:destroy(roomOff)
	end

	self._queues = {}	
end

--[[
	Registers a proxy command with the base driver's 
	PRX_CMD table.
--]]
function AudioQueueManager:registerProxyCommands()

	-- Create a late init method that updates queue settings
	function ON_DRIVER_LATEINIT.audioQueueManagerLateInit()
		Log:Debug("ON_DRIVER_LATEINIT.audioQueueManagerLateInit()")
		local queueSettings = C4:GetDeviceVariable(AudioQueueManager.DigitalAudioDeviceId, AudioQueueManager.QueueSettingsVarId)
		self:updateQueueSettings(queueSettings)
		self:_watchVariables()
	end

	PRX_CMD.AUDIO_URL_SELECTED 			= function(idBinding, tParams) self:AUDIO_URL_SELECTED(tParams) 	end
	PRX_CMD.QUEUE_STATE_CHANGED 		= function(idBinding, tParams) self:QUEUE_STATE_CHANGED(tParams) 	 	end
	PRX_CMD.QUEUE_DELETED 				= function(idBinding, tParams) self:QUEUE_DELETED(tParams) 			 	end
	PRX_CMD.QUEUE_MEDIA_INFO_UPDATED 	= function(idBinding, tParams) self:QUEUE_MEDIA_INFO_UPDATED(tParams) 	end
	PRX_CMD.QUEUE_NEED_NEXT 			= function(idBinding, tParams) self:QUEUE_NEED_NEXT(tParams) 			end
	PRX_CMD.SKIP_FWD 					= function(idBinding, tParams) self:SKIP_FWD(tParams) 				 	end
	PRX_CMD.SKIP_REV 					= function(idBinding, tParams) self:SKIP_REV(tParams) 				 	end
	PRX_CMD.GetQueue 					= function(idBinding, tParams) self:GetQueue(tParams) 				 	end
	PRX_CMD.PlayItem 					= function(idBinding, tParams) self:PlayItem(tParams) 				 	end
	PRX_CMD.GetDashboard 				= function(idBinding, tParams) self:GetDashboard(tParams) 			 	end
	PRX_CMD.ToggleShuffle				= function(idBinding, tParams) self:ToggleShuffle(tParams)			 	end
	PRX_CMD.ToggleRepeat				= function(idBinding, tParams) self:ToggleRepeat(tParams)			 	end
	PRX_CMD.RemoveFromQueue				= function(idBinding, tParams) self:RemoveFromQueue(tParams)		 	end
	PRX_CMD.PlayNextInQueue				= function(idBinding, tParams) self:PlayNextInQueue(tParams)		 	end
end

--[[
	Gets the shuffle and repeat state for the a given room. The state comes from
	the digital audio driver and can be set through programming. AudioQueueManager
	monitors queue state info and updates itself accordingly.

	@param roomId The id of the room to get the settings for.
	@return Returns a table with values repeatVal and shuffleVal
--]]
function AudioQueueManager:getQueueSettingsByRoom(roomId)
	local settings = self._roomSettings[roomId]

	if not settings then 
		return {repeatVal = false, shuffleVal = false}
	end

	return settings
end

--[[
	Sets the shuffle state for a given room. This also sets the room in digital audio.
	@param roomId number. The id of the room to set the shuffle for.
	@param shuffle boolean. The shuffle state.
--]]
function AudioQueueManager:setShuffleForRoom(roomId, shuffleVal)
	local val = 0
	if shuffleVal then
		val = 1
	end

	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "SHUFFLE", {ROOM_ID = roomId, SHUFFLE = val})
end

--[[
	Sets the repeat state for a given room. This also sets the room in digital audio.
	@param roomId number. The id of the room to set the shuffle for.
	@param shuffle boolean. The shuffle state.
--]]
function AudioQueueManager:setRepeatForRoom(roomId, repeatVal)
	local val = 0
	if repeatVal then
		val = 1
	end

	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "REPEAT", {ROOM_ID = roomId, REPEAT = val})
end

--------------------------------------------------------------------------------------------
--								  		UI METHODS									  	  --
--------------------------------------------------------------------------------------------

--[[
	Gets the currently playing queue for the room the navigator is in.
	Note the capitalization is required because this a builtin command
	that comes from the UIs themselves.
--]]
function AudioQueueManager:GetQueue(tParams)
	local roomId = tParams.ROOMID

	local queue = self:getQueueForRoom(roomId)

	-- Inform the queue that it should dispatch a queueChanged event to 
	-- this navigator.
	if queue then
		queue:queueChanged(tParams.NAVID)
	end
end

--[[
	Called to play an item in the queue.
	Note: that the id passed in is NOT the track id. Instead
	it is a unique number assigned to every queue item. This is 
	to allow selecting a track who's track id appears in the queue multiple
	times so that queue track position is preserved.
--]]
function AudioQueueManager:PlayItem(tParams)
	local roomId = tParams.ROOMID
	local queue = self:getQueueForRoom(roomId)

	Log:Trace("AudioQueueManager:PlayItem room:", roomId, " has queue:", (queue ~= nil))
	local args = self:_extractArgsFromXML(tParams.ARGS)

	if queue and args.Id then 
		queue:playItemById(args.Id)
	end
end

function AudioQueueManager:_extractArgsFromXML(argsXML)
	local args = {}
	local parsedArgs = C4:ParseXml(argsXML)
	for i,v in pairs(parsedArgs.ChildNodes or {}) do		
		args[v.Attributes["name"]] = v.Value
	end
	return args
end

--[[
	Called by the proxy when a UI requests the dashboard.
--]]
function AudioQueueManager:GetDashboard(tParams)
	local roomId = tParams.ROOMID

	local queue = self:getQueueForRoom(roomId)

	-- Inform the queue that it should dispatch a queueChanged event to 
	-- this navigator.
	if queue then
		queue:dashboardChanged(tParams.NAVID)
	end
end

function AudioQueueManager:ToggleShuffle(tParams)
	local roomId = tParams.ROOMID
	local queue = self:getQueueForRoom(roomId)

	if queue then 
		queue:toggleShuffle()
	end
end


function AudioQueueManager:ToggleRepeat(tParams)
	local roomId = tParams.ROOMID
	local queue = self:getQueueForRoom(roomId)

	if queue then 
		queue:toggleRepeat()
	end
end

function AudioQueueManager:RemoveFromQueue(tParams)
	local roomId = tParams.ROOMID
	local queue = self:getQueueForRoom(roomId)

	if queue then 
		local args = self:_extractArgsFromXML(tParams.ARGS)
		queue:removeAudioItemById(args.id)
	end
end

function AudioQueueManager:PlayNextInQueue(tParams)
	local roomId = tParams.ROOMID
	local queue = self:getQueueForRoom(roomId)

	if queue and queue:isa(RoomAudioQueue) then 
		local args = self:_extractArgsFromXML(tParams.ARGS)
		queue:moveToNextById(args.id)
	end
end

--------------------------------------------------------------------------------------------
--								  	END	UI METHODS									  	  --
--------------------------------------------------------------------------------------------

--------------------------------------------------------------------------------------------
--								  	PROXY METHODS									  	  --
--------------------------------------------------------------------------------------------

--[[
	Called when the digital audio queues room map variable changes. 

	<?xml version="1.0" encoding="UTF-8"?>
	<info>
		<audioQueueInfo>
			<queue>
				<id>10019</id>
				<name>Living</name>
				<state>Play</state>
				<song id="" source="5"/>
				<next id="" source="5"/>
				<index>0</index>
				<rooms>
					<id>8</id>
					<id>9</id>
				</rooms>
			</queue>
			<roomInfo>
				<roomId>8</roomId>
				<ip>192.168.1.21</ip>
				<port>6203</port>
				<channel>0</channel>
			</roomInfo>
			<roomInfo>
				<roomId>9</roomId>
				<ip>192.168.1.21</ip>
				<port>6200</port>
				<channel>0</channel>
			</roomInfo>
		</audioQueueInfo>
	</info>
--]]
function AudioQueueManager:updateRoomMap(audioQueueInfoXml)
	Log:Debug("AudioQueueManager:updateRoomMap")
	Log:Trace(audioQueueInfoXml)

	local xml = Xml:new(audioQueueInfoXml)
		
	for audioQueueNode in xml:nodesWithName("audioQueueInfo") do
		local queueNode = audioQueueNode:findNode("queue")

		if not queueNode then
			error("AudioQueueManager:updateRoomMap. Could Not find queue node")
		end

		local queueId = queueNode:findNode("id").Value		
		local roomsNode = queueNode:findNode("rooms")

		local rooms = {}
		for idNode in roomsNode:nodesWithName("id") do
			local roomId = idNode.Value
			rooms[#rooms + 1] = roomId			
		end

		self._queueRoomMap[queueId] = rooms

		local queue = self:getQueueById(queueId) 

		-- The queue might be for a different driver playing somewhere else.
		if queue then
			queue:setRoomIds(rooms)			
		end

	end
end

--[[
<room_queue_settings>
	<room_info>
		<roomid>6</roomid>
		<shuffle>0</shuffle>
		<repeat>0</repeat>
	</room_info>
	<room_info>
		<roomid>11</roomid>
		<shuffle>0</shuffle>
		<repeat>0</repeat>
	</room_info>
</room_queue_settings>
--]]
function AudioQueueManager:updateQueueSettings(queueSettingsXml)
	Log:Debug("AudioQueueManager:updateQueueSettings()")
	Log:Trace(queueSettingsXml)

	if not queueSettingsXml or queueSettingsXml == "" then
		Log:Debug("AudioQueueManager:updateQueueSettings() called without queueSettings")
		return
	end

	local xml = Xml:new(queueSettingsXml)

	for roomInfoNode in xml:nodesWithName("room_info") do
		local roomId = 0
		local roomInfo = {}
		for node in roomInfoNode:children() do						
			local name = node.Name
			local val  = node.Value

			if "roomid" == name then
				roomId = val
				roomInfo.roomId = roomId

			elseif "shuffle" == name then
				roomInfo.shuffleVal = toboolean(val)

			elseif "repeat" == name then
				roomInfo.repeatVal = toboolean(val)
			end
		end
		self._roomSettings[roomId] = roomInfo
		self:dispatchEvent(AudioQueueManager.EVENTS.RoomSettingsChanged, roomInfo)
	end
end

--[[
	Called when we receive from the proxy an INTERNET_RADIO_SELECTED.
--]]
function AudioQueueManager:AUDIO_URL_SELECTED(tParams)
	local queueId = tParams["QUEUE_ID"]
	local roomId = tParams["ROOM_ID"]

	Log:Trace("AudioQueueManager: settings queue id: ", queueId, " to room:", roomId)

	-- Set the room's queue's id
	local queue = self:getQueueForRoom(roomId)
	if queue then
		queue:setQueueId(queueId)

		-- Check if we have the rooms for the queue id and set them
		if self._queueRoomMap[queueId] then
			queue:setRoomIds(self._queueRoomMap[queueId])
		end

		queue:AUDIO_URL_SELECTED(tParams)
	else
		Log:Warn("Could not find queue for room:", roomId)
	end
end

function AudioQueueManager:QUEUE_STATE_CHANGED(tParams)
	Log:Trace("AudioQueueManager:QUEUE_STATE_CHANGED")

	local queue = self:getQueueById(tParams.QUEUE_ID)

	if queue then
		queue:QUEUE_STATE_CHANGED(tParams)
	else
		Log:Warn("AudioQueueManager: Could not find queue with id:", queueId)
	end
end

function AudioQueueManager:QUEUE_DELETED(tParams)
	local queueId = tParams.QUEUE_ID

	self._queueRoomMap[queueId] = nil

	for i, queue in ipairs(self._queues) do
		if queue:getQueueId() == queueId then
			queue:destroy()
			table.remove(self._queues, i)
			-- Clear out the rooms
			for i, roomId in ipairs(queue:getRoomIds()) do
				Log:Trace("AudioQueueManager:QUEUE_DELETED removing queue from room", roomId)
			end
			return 
		end
	end
end

function AudioQueueManager:QUEUE_NEED_NEXT(tParams)
	Log:Trace("AudioQueueManager:QUEUE_NEED_NEXT")

	local queue = self:getQueueById(tParams.QUEUE_ID)

	if queue then
		queue:QUEUE_NEED_NEXT(tParams)
	else
		Log:Warn("AudioQueueManager: Could not find queue with id:", queueId)
	end
end

function AudioQueueManager:SKIP_FWD(tParams)
	local queue = self:getQueueForRoom(tParams.ROOM_ID)

	if queue then
		queue:SKIP_FWD()
	else
		Log:Warn("AudioQueueManager:SKIP_FWD No known queue for room:", tParams.ROOM_ID)
	end
end

function AudioQueueManager:SKIP_REV(tParams)
	local queue = self:getQueueForRoom(tParams.ROOM_ID)

	if queue then
		queue:SKIP_REV()
	else
		Log:Warn("AudioQueueManager:SKIP_REV No known queue for room:", tParams.ROOM_ID)
	end
end

--[[
MEDIA_INFO:  <mediainfo>
	<deviceid>100002</deviceid>
	<queueid>10001</queueid>
	<streamid>0</streamid>
	<streamStatus>status=playing,ts=323242,ex=321134</streamStatus>
	<source>5</source>
	<mediatype>SONG</mediatype>
	<mediatypeV2>GENERIC_MEDIA</mediatypeV2>
	<medSrcDev>1574</medSrcDev>
	<stationid>rhap://Tra.136181759</stationid>
	<queueInfo>0</queueInfo>
	<album>Love, Marriage??? &amp; Divorce</album>
	<artist>Toni Braxton</artist>
	<title>I Wish</title>
	<channel/>
	<img>aHR0cDovL3N0YXRpYy5yaGFwLmNvbS9pbWcvNjAweDYwMC81LzYvMC8xLzQ5NDEwNjVfNjAweDYwMC5qcGc=</img>
</mediainfo>
--]]
function AudioQueueManager:QUEUE_MEDIA_INFO_UPDATED(tParams)
	local queue = self:getQueueById(tParams.QUEUE_ID)

	if queue then
		local infoXml = Xml:new(tParams.MEDIA_INFO)
		local streamNode = infoXml:findNode("streamid")
		
		if streamNode then
			queue:setStreamId(streamNode.Value)
		end
		
		local streamStatusNode = infoXml:findNode("streamStatus")
		if streamStatusNode then
			queue:setStreamStatus(streamStatusNode.Value)
		end

	end
end

--------------------------------------------------------------------------------------------
--	    						 END PROXY METHODS					  					  --
--------------------------------------------------------------------------------------------

