--------------------------------------------------------------------------------------------
--	BaseAudioQueue.lua
--
--	Represents an audio queue of data that the digital audio server will
--  play. Each queue is associated with at least one room.
--
--  Copyright 2014 Control4 Corporation. All rights reserved.
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"
require "law.Xml"
require "law.StringUtils"


BaseAudioQueue = inheritsFrom(nil)

BaseAudioQueue.States = {
	Stopped = 0, -- The queue has stopped playing
	Playing = 1, -- The queue is currently playing
	Paused  = 2, -- The queue is paused.
	Deleted = 3, -- All done with this queue
	Empty = 4, -- No items in the queue.
}

-- The number of errors that occur inside the ErrorTimeout that triggers
-- the queue to stop playing.
BaseAudioQueue.MaxErrorRate = 5

-- The how often the error timeout should fire to check if the total
-- number of errors in that time period exceeds MaxErrorRate.
BaseAudioQueue.ErrorTimeout = 5000

function BaseAudioQueue:init(roomId, proxyBindingId)
	Log:Debug("BaseAudioQueue:init, self:", tostring(self) )
	-- Default audio room, Room that first created this queue
	self._rooms = {roomId}
	self._proxyBindingId = proxyBindingId
	-- Set the queue to the stopped state.
	self._state = self.States.Stopped
	-- Set a dummy queue id.
	self._queueId = -1

	-- The os.time() value of when the last played track started.
	self._audioItemStartTime = -1

	self._errorCount = 0
end

--[[
	Copies settings from a queue to this one.
--]]
function BaseAudioQueue:copy(queue)
	self._rooms = deepcopy(queue._rooms)
	self._proxyBindingId = queue._proxyBindingId
	self._state = queue._state
	self._queueId = queue._queueId
end

--[[
	Called when the the queue is being deleted.
	@param If true all rooms should be turned off.
--]]
function BaseAudioQueue:destroy(roomOff)
	self._state = self.States.Deleted
	self:dashboardChanged()

	if roomOff and self._queueId ~= -1 then
		local tParams = {}
    	tParams.QUEUE_ID = self._queueId
    	tParams.ROOM_ID = self:getOwningRoom()
    	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "REMOVE_QUEUE", tParams)
    end
end

--[[
	Sets the digital audio assigned queueId
--]]
function BaseAudioQueue:setQueueId(queueId)
	Log:Trace("BaseAudioQueue queue id set for rooms", unpack(self._rooms))
	self._queueId = queueId
end

function BaseAudioQueue:getQueueId()
	return self._queueId
end

--[[
	Sets the digital audio stream associated with the queue
--]]
function BaseAudioQueue:setStreamId(streamId)
	Log:Trace("BaseAudioQueue stream id set for rooms", unpack(self._rooms))
	self._streamId = streamId
end

function BaseAudioQueue:getStreamId()
	return self._streamId
end

--[[
	Sets the stream status associated with the queue.
--]]
function BaseAudioQueue:setStreamStatus(status)
	Log:Trace("BaseAudioQueue stream status id set for rooms", unpack(self._rooms))
	self._streamStatus = status
end

function BaseAudioQueue:getStreamStatus()
	return self._streamStatus
end



--[[
	@param roomId The id of the room to add to the zone.
--]]
function BaseAudioQueue:addRoom(roomId)
	-- Check if the room is in the queue.
	for i, v in ipairs(self._rooms) do
		if roomId == v then
			return
		end
	end

	table.insert(self._rooms, roomId)
end

--[[
	Removes a room from the zone.
	@param roomId The id of the room to remove from the zone.
--]]
function BaseAudioQueue:removeRoom(roomId)
	for i, v in ipairs(self._rooms) do
		if roomId == v then
			table.remove(self._rooms, i)
			return
		end
	end
end

function BaseAudioQueue:isInRoom(roomId)
	for i,v in ipairs(self._rooms) do
		if roomId == v then
			return true
		end
	end

	return false
end

function BaseAudioQueue:setRoomIds(rooms)
	self._rooms = rooms
end

--[[
	Gets the current roomIds as an array. The array is a copy of the original.
--]]
function BaseAudioQueue:getRoomIds()
	return deepcopy(self._rooms)
end

--[[
	Gets the room this queue was created for.
	It is assumed to be the first one in the list.
--]]
function BaseAudioQueue:getOwningRoom()
	return self._rooms[1]
end

--[[
	Virtual method should be overriden by sub-class.
	Gets the audio items in the queue. The default
	behavior is to return an empty list.
--]]
function BaseAudioQueue:getQueueItems()
	return {}
end

--[[
	Gets the MSP proxy binding id for the queue
--]]
function BaseAudioQueue:getProxyBinding()
	return self._proxyBindingId
end

function BaseAudioQueue:setProxyBinding(bindingId)
	self._proxyBindingId = bindingId
end

--[[
	The number of seconds since the currently playing item
	in the queue started playing.
--]]
function BaseAudioQueue:secondsSinceLastPlay()
	return os.difftime(os.time(), self._audioItemStartTime)
end

--[[
	Plays an item identified by the items id.
	@param id The queues id. Not a trackId.

	@note trackId is not used becuase multiple items in the
	queue could have the same track id.
--]]
function BaseAudioQueue:playItemById(id)
	Log:Debug("BaseAudioQueue:playItemById()")
	error("BaseAudioQueue:playItemById() Pure Virtual - THis must be implemented by sub classes")
end


--[[
	plays a specific autio item
	@param item An AudioQueueItem
--]]
function BaseAudioQueue:_playItem(item, nextItem)
  --Log:Debug("playing:", item:getUrl())
  Log:Debug("BaseAudioQueue:_playItem(), Playing:", item:getTitle(), " in room:", self._rooms[1],  " with station:", item:getUrl())
  local radioParams = {}
  radioParams["ROOM_ID"] = self._rooms[1]
  radioParams["STATION_URL"] = item:getUrl()
  if nextItem ~= nil then
    radioParams["NEXT_URL"] = nextItem:getUrl()
  end
  self._audioItemStartTime = os.time() -- Get the time the song started. Used for skip back functionality.
  radioParams["QUEUE_INFO"] = self._audioItemStartTime
  radioParams["VOLUME"] = self._volume
  self._volume = nil
  C4:SendToProxy(self._proxyBindingId, "SELECT_AUDIO_URL", radioParams, "COMMAND")

  self:_updateMediaInfo(item)
end

function BaseAudioQueue:_updateMediaInfo(item)
  for i,roomId in ipairs(self._rooms) do
	  local params = {}
	  params["ROOMID"] = roomId
	  params["IMAGEURL"] = item:findLargestImageUrl()

	  params["LINE1"] = item:getTitle()
	  params["LINE2"] = item:getArtist()
	  params["LINE3"] = item:getAlbum()
	  params["LINE4"] = ""
	  params["MEDIATYPE"] = "secondary"
	  --params["QUEUEID"] = queueId
	  --params["MERGE"] = "True"
	  params["FORCE"] = "1"
	  C4:SendToProxy(self._proxyBindingId, "UPDATE_MEDIA_INFO", params, "COMMAND", true)
  end
end

function BaseAudioQueue:_setNextItem(item)
	local nextParams = {}
	nextParams["ROOM_ID"] = self._rooms[1]
	if item ~= nil then
		Log:Debug("BaseAudioQueue:_setNextItem(), Play next:", item:getTitle(), " with station:", item:getUrl())
		nextParams["NEXT_URL"] = item:getUrl()
	else
		Log:Debug("BaseAudioQueue:_setNextItem(), clear next")
		nextParams["NEXT_URL"] = ""
	end
	C4:SendToProxy(self._proxyBindingId, "SET_NEXT_AUDIO_URL", nextParams, "COMMAND")
end

--[[
	Clears all of the media info.
--]]
function BaseAudioQueue:clearQueue()
	self._state = self.States.Empty

	self:dashboardChanged()

	for i,roomId in ipairs(self._rooms) do
	  local params = {}
	  params["ROOMID"] = roomId
	  params["IMAGEURL"] = ""

	  params["LINE1"] = ""
	  params["LINE2"] = ""
	  params["LINE3"] = ""
	  params["LINE4"] = ""
	  params["MEDIATYPE"] = "secondary"
	  --params["QUEUEID"] = queueId
	  --params["MERGE"] = "True"
	  params["FORCE"] = "1"
	  C4:SendToProxy(self._proxyBindingId, "UPDATE_MEDIA_INFO", params, "COMMAND", true)
  end

  -- Note there is a race condition where the MedSrcDev id is being cleared when this is called.
  -- This call must come after the UPDATE_MEDIA_INFO
  C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "CLEAR_QUEUE", {ROOM_ID=self:getOwningRoom()})
end

--[[
	Starts playing the queue from a stopped or paused state. This might
	notify the UI that the now playing has changed.
--]]
function BaseAudioQueue:play()
	Log:Debug("BaseAudioQueue:play()")
	error("BaseAudioQueue:play() Pure Virtual - This must be implemented by sub classes.")
end

--[[
	Puts the queue in the paused state.
--]]
function BaseAudioQueue:pause()
	Log:Debug("BaseAudioQueue:pause()")

	-- You can't go from empty to paused.
	if self._state == self.States.Empty then
		Log:Info("Can't pause, queue is empty")
		return
	end

	self._state = self.States.Paused
	self:dashboardChanged()
end

--[[
	Puts the queue in the stopped state. Resuming playing
	will result in the queue restarting from the beginning.
--]]
function BaseAudioQueue:stop()
	Log:Debug("BaseAudioQueue:stop()")

	-- You can't go from empty to stopped.
	if self._state == self.States.Empty then
		Log:Info("Can't stop, queue is empty")
		return
	end

	self._state = self.States.Stopped
	self:dashboardChanged()
end

--[[
	Called when the end of the currently playing item has been reached.
--]]
function BaseAudioQueue:endOfItem(queueInfo)
	Log:Debug("BaseAudioQueue:endOfItem()")

	-- Sometimes we don't have the queue info.
	if queueInfo == nil then
		return
	end

	local timePlaying = os.time() - tonumber(queueInfo)

	if timePlaying <= 2 then
		Log:Debug("BaseAudioQueue: Last item played for less than or equal to 2 seconds.")
		-- This track has played less than 1 second. This is probably due to an error.
		self._errorCount = self._errorCount + 1

		-- Track the number of errors that happen inside of a certain time period.
		if self._errorTimer == nil then
			self._errorTimer = ServiceLocator.TimerManager:addTimer(BaseAudioQueue.ErrorTimeout)

			self._errorTimer:done(function()
				Log:Debug("BaseAudioQueue errorTimer:done()")
				self._errorTimer = nil;
				if self._errorCount >= BaseAudioQueue.MaxErrorRate then
					self:stop()
				else
					self._errorCount = 0;
				end
			end)
		end
	end

end

--[[
	Sends the QueueChanged event when the List of queue items has changed. Don't
	call this method if only updating the NowPlaying or NowPlayingIndex data.

	@param navId [Optional] Send the event to a specific navigator.
			If left off the event is sent to all of the rooms that belong
			to the zone.
--]]
function BaseAudioQueue:queueChanged(navId)
	Log:Debug("BaseAudioQueue:queueChanged")
	error("BaseAudioQueue:queueChanged() - Pure Virtual - This must be implemented by sub classes.")
end

--[[
	Call when the currently playing track index changes.
--]]
function BaseAudioQueue:nowPlayingChanged()
	Log:Debug("BaseAudioQueue:nowPlayingChanged")
	error("BaseAudioQueue:nowPlayingChanged() - Pure Virtual - This must be implemented by sub classes.")
end

--[[
	Call when the global data changes for actions.
--]]
function BaseAudioQueue:globalDataChanged()
	Log:Debug("BaseAudioQueue:globalDataChanged")
	error("BaseAudioQueue:globalDataChanged() - Pure Virtual - This must be implemented by sub classes.")
end

--[[
	Updates the queues state based on the stateName passed into
	the method

	@param A caseinsensitive state name. Possible values are:
		PLAY, PAUSE, STOP, END

	@param queueInfo [OPTIONAL] The queue info param from the queue. In these queues it holds a timestamp.
--]]
function BaseAudioQueue:updateState(stateName, queueInfo)
	local state = stateName:upper()

	if "PLAY" == state then
		self:play()

	elseif "PAUSE" == state then
		self:pause()

	elseif "STOP" == state then
		self:stop()

	elseif "END" == state then
		self:endOfItem(queueInfo)
	end
end

--[[
	May be overridden to supply information about the following track (gapless playback)

	@param queueInfo [OPTIONAL] The queue info param from the queue. In these queues it holds a timestamp.
--]]
function BaseAudioQueue:needNext(queueInfo)
	Log:Debug("BaseAudioQueue:needNext()")
end

--[[
	Sends the dashboard changed event. The possible
	dashboard items are SkipRev, Play, Pause, SkipFwd, ShuffleOn, ShuffleOff, RepeatOn, RepeatOff

	@param navId [Optional] Send the event to a specific navigator.
				  Otherwise the event goes to all rooms in the zone.
--]]
function BaseAudioQueue:dashboardChanged(navId)
	local buttons

	if self._state == self.States.Playing then
		buttons = {"SkipRev", "Pause", "SkipFwd"}
	elseif self._state == self.States.Paused then
		buttons = {"SkipRev", "Play", "SkipFwd"}
	elseif self._state == self.States.Stopped then
		buttons = {"Play"}
	elseif self._state == self.States.Deleted or
	       self._state == self.States.Empty then
		buttons = {}
	else
		Log:Error("BaseAudioQueue:_dashboardChanged() Unknown state")
		return
	end

	if (USE_CUSTOM_DASH) then
		if (self._shuffleMode) then
			table.insert (buttons, 1, 'ShuffleOff')
		else
			table.insert (buttons, 1, 'ShuffleOn')
		end

		if (self._repeatMode) then
			table.insert (buttons, 'RepeatOff')
		else
			table.insert (buttons, 'RepeatOn')
		end
	end

	local tParams = {
		Items = table.concat(buttons, " ")
	}

	-- The Room ids will be added in _sendEvent. That's how navigator's know
	-- which rooms the dashboard applies to.
	self:_sendEvent("DashboardChanged", tParams, navId)
end

--[[
	Sends an event to all rooms that are "attached" to the zone.
--]]
function BaseAudioQueue:_sendEvent(evtType, data, navId, rawDataNames)
	local tParams = {
		NAME = evtType,
		EVTARGS = Xml.BuildSimpleXml(nil, data, true, rawDataNames),
	}

	if navId then
		tParams.NAVID = navId
	else
		tParams.ROOMS = table.concat(self._rooms, ",")
	end

	Log:Trace("BaseAudioQueue:_sendEvent", evtType, tParams)
	C4:SendToProxy(self._proxyBindingId, "SEND_EVENT", tParams, "COMMAND")
end

--------------------------------------------------------------------------------------------
--								  RECEIVED FROM PROXY METHODS						  	  --
--------------------------------------------------------------------------------------------

--[[
	Called by the audio queue manager to indicate the song ended.
--]]
function BaseAudioQueue:QUEUE_STATE_CHANGED(tParams)
	assert(tParams.QUEUE_ID == self:getQueueId(), "BaseAudioQueue:QUEUE_STATE_CHANGED, invalid queue id.")

	self:updateState(tParams.STATE, tParams.QUEUE_INFO)
end

--[[
	Called by the audio queue manager to indicate that we should supply information about the next song
--]]
function BaseAudioQueue:QUEUE_NEED_NEXT(tParams)
	assert(tParams.QUEUE_ID == self:getQueueId(), "BaseAudioQueue:QUEUE_NEED_NEXT, invalid queue id.")

	self:needNext(tParams.QUEUE_INFO)
end

function BaseAudioQueue:SKIP_FWD(tParams)
	Log:Debug("BaseAudioQueue:SKIP_FWD")
end

function BaseAudioQueue:SKIP_REV(tParams)
	Log:Debug("BaseAudioQueue:SKIP_REV")
end

function BaseAudioQueue:AUDIO_URL_SELECTED(tParams)
	Log:Debug("BaseAudioQueue:AUDIO_URL_SELECTED")
end
--------------------------------------------------------------------------------------------
--								  RECEIVED FROM PROXY METHODS						  	  --
--------------------------------------------------------------------------------------------
