--------------------------------------------------------------------------------------------
--	BinarySearchCollection.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
-- A collection of items that implements the binary search patter to find an item in the list
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"

BinarySearchCollection = inheritsFrom(nil)


--[[
	Creates the collection.
	@param collection. An optional array that MUST be in order.
--]]
function BinarySearchCollection:init(collection)
	
	if not collection then 
		self._collection = {}
	else
		self._collection = collection
	end

	self._comparer = function(a,b)
		if a < b then
			return -1		
		elseif a > b then 
			return 1
		else
			-- a == b
			return 0
		end
	end

	-- The last index that was searched for. Used for fast insertions	
	self._lastIndex = 1

	-- Provide an extra layer of collection helpers for direct index access.	
	local mt = 
	{
		__index = BinarySearchCollection._index,
		__metatable = "Protected"
	}
	setmetatable(mt, BinarySearchCollection)
	setmetatable(self, mt)
end

function BinarySearchCollection._index(tBSC, key)
	--print(tBSC, key)
	-- Do standard stuff
	if BinarySearchCollection[key] then
		--print("bsc has key", key)
		return BinarySearchCollection[key]
	end

	-- Get the _collection object without invoking this function
	local collection = rawget(tBSC, "_collection")
	
	if not collection then 
		return nil
	end
	
	return collection[key]
end

--[[
	Gets the raw collection array. 
--]]
function BinarySearchCollection:rawget()
	return self._collection
end	

--[[
	Sets the function used to compare 2 items. The function should
	have the following signature: function(a, b) and return one of the
	following values:
	0 : a == b, -1: a < b. 1: a > b.

	@params comparer. The comparison function
--]]
function BinarySearchCollection:setCompareFunction(comparer)
	self._comparer = comparer
end

function BinarySearchCollection:length()
	return #self._collection
end

function BinarySearchCollection:hasItem(item)	
	return toboolean(self:findItemIndex(item))
end

--[[
	Inserts an item if it is not already in the collection
--]]
function BinarySearchCollection:insert(item, usePreviousSearchIndex)
	local len = self:length()

	if len == 0 then
		self._lastIndex = 1

	elseif not usePreviousSearchIndex and self:findItemIndex(item) then
		Log:Trace("BinarySearchCollection already has item. Not inserting", item)
		return false
	end


	if self[self._lastIndex] and self._comparer(self[self._lastIndex], item) < 0 then
		self._lastIndex = self._lastIndex + 1
	end

	--print("inserting item", item, "at:", self._lastIndex)
	table.insert(self._collection, self._lastIndex, item)	
	return true
end

--[[
	Removes an item from the table
--]]
function BinarySearchCollection:remove(item, usePreviousSearchIndex)
	if not usePreviousSearchIndex and not self:findItemIndex(item) then
		Log:Debug("BinarySearchCollection does not have the item. Not removing")
		return false
	end

	table.remove(self._collection, self._lastIndex)
	return true
end

--[[
	Gets an iterator that returns the next value in the list. 
	Note modifying the list invalidates the iterator
--]]
function BinarySearchCollection:iter()
	local i = 0
	return function()	
		i = i + 1	
		return self[i]
	end
end

function BinarySearchCollection:findItemIndex(item)
	if self:length() == 0 then
		self._lastIndex = 1
		return nil
	end

	-- For proper tail recursion the search must be last
	return self:_search(item, 1, self:length())
end

function BinarySearchCollection:_search(item, startIndex, endIndex)

	if (startIndex > endIndex) then
		self._lastIndex = startIndex
		return nil
	end

	local mid = math.floor((startIndex + endIndex) / 2)

	local comparison = self._comparer(self[mid] , item)

	if comparison < 0 then
		startIndex = mid + 1
		self._lastIndex = startIndex

	elseif comparison > 0 then
		endIndex = mid-1
		self._lastIndex = endIndex

	else
		self._lastIndex = mid
		return mid
	end

	-- For proper tail recursion the search must be last.
	return self:_search(item, startIndex, endIndex)
end

function BinarySearchCollection:dump()
	for v in self:iter() do
		print(v)
	end
end

--[[
	Test the binary search collection
	@param log True if the test should print out.
--]]
function BinarySearchCollection.test(log)
	if log then
		print("Startings bsc test")
	end

	local bsc = BinarySearchCollection:new()
	
	assert(bsc:length() == 0, "length is not 0.")

	assert(not bsc:hasItem(1), "bsc:hasItem(1) is true when it should be false")
	assert(bsc:insert(1), "bsc:insert(1) could not insert")

	assert(bsc[1] == 1, "bsc[1] == 1 is false")
	assert(bsc:length() == 1, "length is not 1.")

	assert(bsc:remove(1), "Cannot remove 1")
	assert(bsc:length() == 0, "length is not 0")

	local len = 500
	for i = 1, len do
		bsc:insert(i)
		assert(bsc[i] == i, "Inserted item is not at index")
		assert(bsc:length() == i, "bsc:length() is not equal to " .. i)
	end	

	assert(bsc:length() == len, "Invalid len")

	for i = len, 2, -1 do		
		assert(bsc[i] > bsc[i-1], "bsc[i] > bsc[i-1], is not true")
	end

	local lastIndex = 0
	local i = 0
	for v in bsc:iter() do
		i = i + 1
		lastIndex = i
		assert(i == v, "The index does not equal the value. Something is wrong with ipairs. Index: " .. i .. " val: " .. v)
		assert(i == bsc:findItemIndex(v), "findItemIndex is busted, i =" .. i)
	end

	assert(lastIndex == len, "iterator did not iterate over entire collection. lastIndex: " .. lastIndex)

	bsc = BinarySearchCollection:new()

	for i = 1, len do
	 	bsc:insert(math.random(1,1000))
	end
		
	for i = bsc:length(), 2, -1 do		
		assert(bsc[i] > bsc[i-1], "bsc[i] > bsc[i-1], is not true")
	end

	bsc = BinarySearchCollection:new({1, 5, 6, 20, 35})
	assert(bsc:hasItem(1), "Cannot find 1")
	assert(bsc:hasItem(5), "Cannot find 5")
	assert(bsc:hasItem(6), "Cannot find 6")
	assert(bsc:hasItem(20), "Cannot find 20")

	bsc = BinarySearchCollection:new({{val = 1}, {val = 2}, {val = 3}})

	bsc:setCompareFunction(function(a,b)
		if a.val < b.val then
			return -1
		elseif a.val > b.val then
			return 1
		else
			return 0
		end
	end)

	assert(bsc:hasItem({val = 1}), "Cannot find {val = 1}")
	assert(bsc:hasItem({val = 2}), "Cannot find {val = 2}")
	assert(bsc:hasItem({val = 3}), "Cannot find {val = 3}")
	assert(not bsc:hasItem({val = 4}), "Should NOT find {val = 4}")

	bsc:insert({val = 2.5})
	assert(bsc:hasItem({val = 2.5}), "Cannot find {val = 2.5}")

	assert(#bsc:rawget() == 4, "rawget() not correct length")

	if log then
		print("bsc test Done")
	end
end



