--------------------------------------------------------------------------------------------
--	Debugger.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Implements a lua debugger that can be used to remotely debug a lua script
--------------------------------------------------------------------------------------------

json = require "json"
socket = require "socket"

require "law.Log"
require "law.BinarySearchCollection"
require "law.TimerPromise"
require "law.ServiceLocator"
require "law.TimerManager"


Debugger = inheritsFrom(nil)

Debugger.READ_TIMEOUT = 1000

--[[
	Creates the debugger on a port. 
--]]
function Debugger:init(port, noReadTimer)
	-- Check that the main library has been initialized:
	if not ServiceLocator.TimerManager and not noReadTimer then
		-- true adds to ServiceLocator
		TimerManager:new(true)
	end

	self._port = port

	self._noReadTimer = noReadTimer

	self._lastAccessedVar = nil

	self:reset()
end

--[[
	Resets the object for another debug session
--]]
function Debugger:reset()
	if self._client ~= nil then 
		pcall(self._client.close, self._client)
		self._client = nil
	end

	if self._socket ~= nil then
		pcall(self._socket.close, self._socket)
		self._socket:close()
		self._socket = nil
	end

	self._breakFunc = nil
	self:_setupBreakpointCollection()
	self:clearFunctionCache()
end

--[[
	Clears out all of the breakpoints. This should be called
	with clearFunctionCahche()
--]]
function Debugger:_setupBreakpointCollection()
	self._breakpoints = BinarySearchCollection:new()

	self._breakpoints:setCompareFunction(function(a, b)
		if a.source < b.source then 
			return -1
		elseif a.source > b.source then
			return 1
		else
			-- Files are equal check the line number
			return a.currentline - b.currentline			
		end		
	end)
end

function Debugger:clearFunctionCache()
	-- Table of functions that stores a list of functions 
	-- and if the breakpoints are in a given function
	self._functions = {}
	-- Mode k for weak keys instead of values (v is for values)
	setmetatable(self._functions, {__mode = "k"}) 
end

--[[
	Sets a breakpoint on a line in a file.
	@param file The name of the file. On a control4 system this should be the full path to the file.
	@param line the file's line number.
--]]
function Debugger:setBreakpoint(file, line, clear)
	-- Changes to breakpoints means we need to rebuild the cache. 
	-- See http://lua-users.org/lists/lua-l/2007-03/msg00016.html
	self:clearFunctionCache()

	if not clear then
		print("Setting breakpoint:", line, file)
	else
		print("Clearing breakpoint", line, file)
	end

	-- comp is a field used to easily compare breakpoints
	local bp = {source = "@".. file, currentline = line}

	if not clear and not self._breakpoints:hasItem(bp) then
		self._breakpoints:insert(bp, true)
	elseif clear then
		self._breakpoints:remove(bp)
	end

	-- See if there are any more breakpoints or commands
	self:_checkSocket()
end

--[[
	Starts the debugger.
--]]
function Debugger:start()
	self._running = true
	print("C4 Debugger: Starting session...")
	-- Listen for a connection.
	--self._socket = socket.bind("localhost", self._port)
	
	-- Get a master socket
	self._socket = socket.tcp()

	self._socket:bind(C4:GetControllerNetworkAddress(), self._port)

	-- Turns socket into server
	self._socket:listen(1)
	self._socket:setoption("reuseaddr", true)
	
	-- Wait for a connection
	self._client = self._socket:accept()

	self:_checkSocket()

	if not self._noReadTimer then
		self:_setupReadTimeout()
	end

	self:_setupHook()
end

--[[
	Stops the debugger.
--]]
function Debugger:stop()
	debug.sethook()

	self:reset()

	self._running = false

	print("C4 Debugger: Session ended")
end

function Debugger:_checkSocket()
	if not self._client then
		return
	end

	local read, write, err = socket.select({self._client}, nil, .3)
	
	--print(#read, #write, err)
	if read and #read > 0 then
		self:_receive()
	end

	if not self._noReadTimer then
		self:_setupReadTimeout()
	end
end

function Debugger:_setupReadTimeout()
	self._readTimer = ServiceLocator.TimerManager:addTimer(self.READ_TIMEOUT)

	self._readTimer:done(function()
		self:_checkSocket()
	end)
end



function Debugger:_setupHook()
	if not self._running then 
		print("Debugger: setupHook not running")
		return
	end

	local function breakpointHook(evt, line)		
		--print(evt, line)
		if evt == "call" then
			self:_checkFuncForBreakpoint(3)
		elseif evt == "return" or evt == "tail return" then
			self:_checkFuncForBreakpoint(4) -- Check the function we are returning to
		elseif "line" == evt then
			local info = debug.getinfo(2, "Sl")
			--print("on line", info.currentline, info.source)
			if self._breakpoints:hasItem(info) then
				-- Send out the evt and wait for a response
				self._breakFunc = debug.getinfo(2,"f").func
				self:_sendAndEval({t = "b", info = debug.getinfo(2, "Snlu")})
			end
		else 
			print("Debugger._setupHook() unknown evt type:", evt)
		end
	end

	--print("Debugger setting hook to:", breakpointHook)
	if not self._breakFunc and not self._functions[self._breakFunc] then
		--print("set hook cr")
		debug.sethook(breakpointHook, "cr")
	else
		--print("set hook crl")
		debug.sethook(breakpointHook, "crl")
	end
end


function Debugger:_checkFuncForBreakpoint(level)
	local func = debug.getinfo(level, "f")
	--print(func)	

	if func and func.func then
		func = func.func
	else
		local info = debug.getinfo(level, "Sn")
		if not info then 
			return
		elseif info.source == "[C]" then
			print("C Source", info.name) 
			return
		end
	end

	if func and self._functions[func] == nil then

		local info = debug.getinfo(level, "Sn")
		--print("checkFunc:", "name", info.name, "source", info.source, "linedef:", info.linedefined, "lastline:", info.lastlinedefined)
		-- Just in case there aren't any breakpoints.
		self._functions[func] = false

		-- Search for breakpoints in the function
		for i,v in ipairs(self._breakpoints:rawget()) do
			if v.source == info.source and 
			   v.currentline >= info.linedefined and
			   v.currentline <= info.lastlinedefined then
			   -- We found a breakpoint defined in this function
			   --print("Function found with breakpoint", func)
			   self._functions[func] = true
			end
		end
	end

	if func and self._functions[func] == true then
		--print("Setting hook to crl")
		debug.sethook(debug.gethook(), "crl")
	else
		-- turn off the slow debugging.
		--print("Setting hook to cr")
		debug.sethook(debug.gethook(), "cr")			
	end
end

--[[
	Sends an event to the client. Waits for a command response.
--]]
function Debugger:_sendAndEval(tEvt)
	local packet = json:encode(tEvt)
	print("Debugger sending:", packet)
	local lastIndexSent, err, lastErrIndex = self._client:send(packet .. "\n")

	if err then
		print("Debugger._sendAndEval(), socket received error:", err)
	end
	
	self:_receive()
end

function Debugger:_receive()
	local response, err = self._client:receive("*l")

	if response then
		self:_processResponse(response)
	else
		print("Debugger received socket error", err)
		self:stop()
	end
end

--[[
	Processes a response from a breakpoint hit or another command.
--]]
function Debugger:_processResponse(response)
	local tCmd = json:decode(response)

	print("Debugger recv:", response)

	if "n" == tCmd.cmd then
		-- Next command
		self:_doNext()
	elseif "s" == tCmd.cmd then
		-- Step (into) command
		self:_doStep()
	elseif "so" == tCmd.cmd then
		-- Step out command
		self:_doStepOut()
	elseif "e" == tCmd.cmd then
		-- Execute code
	elseif "l" == tCmd.cmd then
		-- List local variables
		self:_listLocals()
	elseif "u" == tCmd.cmd then 
		-- List up values (at level)
	elseif "b" == tCmd.cmd then
		-- set / clear a breakpoint
		self:setBreakpoint(tCmd.filename, tCmd.line, tCmd.clear)		
	elseif "c" == tCmd.cmd then
		-- continue running until next breakpoint
		self:_setupHook()
	elseif "var" == tCmd.cmd then
		self:_drillIntoVar(tCmd.name)
	elseif "clearVar" == tCmd.cmd then
		self:_clearVar()
	elseif "localVar" == tCmd.cmd then
		self:_getLocal(tCmd.name)
	elseif "upVar" == tCmd.cmd then
		self:_getUpVal(tCmd.name)
	elseif "eval" == tCmd.cmd then
		self:_doEval(tCmd.code)
 	else
		print("C4 Debugger: unkown command...running")
	end 
end

--[[
	Steps to the next line skipping over functions.
--]]
function Debugger:_doNext()
	-- The current function level. 
	local level = 0
	local function nextHook(evt, line)
		local info = debug.getinfo(2, "S")
		if evt == "call" and info.linedefined >= 0 then
			level = level + 1
		elseif (evt == "return" or evt == "tail return") and info.linedefined >= 0 then
			level = level - 1 
		elseif evt == "line" and level <= 0 then
			self._breakFunc = debug.getinfo(2,"f").func
			self:_sendAndEval({t = "n", info = debug.getinfo(2, "Snlu")})
		end
	end

	debug.sethook(nextHook, "crl")
end

--[[
	Steps the debugger to the next line and possibly into functions.
--]]
function Debugger:_doStep()
	local function stepHook(evt, line)
		if evt == "line" then
			self._breakFunc = debug.getinfo(2,"f").func
			self:_sendAndEval({t = "s", info = debug.getinfo(2, "Snlu")})
		end
	end

	debug.sethook(stepHook, "l")
end

--[[
	Steps the debugger to the next line and possibly into functions.
--]]
function Debugger:_doStepOut()
	-- The current function level. 
	local level = 0
	local function nextHook(evt, line)
		local info = debug.getinfo(2, "S")
		if evt == "call" and info.linedefined >= 0 then
			level = level + 1
		elseif (evt == "return" or evt == "tail return") and info.linedefined >= 0 then
			level = level - 1 
			if level <= -1 then
				debug.sethook(debug.gethook(), "l")
			end
		elseif evt == "line" and level <= -1 then
			self._breakFunc = debug.getinfo(2,"f").func
			self:_sendAndEval({t = "so", info = debug.getinfo(2, "Snlu")})
		end
	end

	debug.sethook(nextHook, "cr")
end

function Debugger:_listLocals()
	print("Debugger:_listLocals()")

	-- Find the level of the breakFunc. The function
	-- where the debugger is stopped.
	local level = self:_findFuncLevel(1)
	
	local vars = self:_getLocals(level)

	local upVars = self:_getUpVals(level)

	self:_sendAndEval({t = "l", vars = vars, upvars= upVars})
end

function Debugger:_getLocals(level)
	local vars = {}
	do local i = 1; repeat
		local name, value = debug.getlocal(level, i)
		if name then 
			vars[name] = tostring(value)
		end
		i = i + 1
	until name == nil end

	return vars
end

function Debugger:_findFuncLevel(startLvl)
	local level = startLvl

	if self._breakFunc then
		do repeat
			level = level + 1
			local curFunc = debug.getinfo(level, "f").func
		until curFunc == self._breakFunc or curFunc == nil end
		print("Found local function at level:", level )
		return level 
	end

	return nil
end

function Debugger:_getLocal(desiredName)
	local level = self:_findFuncLevel(1) - 1
	local index = 1
	local name = nil
	local value = nil

	do repeat
		name, value = debug.getlocal(level, index)
		index = index + 1
	until name == desiredName or name == nil end

	print("Local variable:", name)
	self:__sendVar(value)	
end

--[[
	Drills into the saved variable if there is one.
--]]
function Debugger:_drillIntoVar(name)
	if not self._lastAccessedVar then
		self:_sendAndEval({t="var", chainEnd=true})
		return
	end

	local var = self._lastAccessedVar[name]
	self:__sendVar(var)
end

function Debugger:_clearVar()
	self._lastAccessedVar = nil
	self:_sendAndEval({t="clearVar"})
end

--[[
	Gets a specific up value.
--]]
function Debugger:_getUpVal(desiredName)
	local level = self:_findFuncLevel(1)
	local func = debug.getinfo(level, "f").func
	local name = nil
	local value = nil
	local index = 1

	do repeat
		name, value = debug.getupvalue(self._breakFunc, index)
		index = index + 1
	until name == desiredName or name == nil end
	self:__sendVar(value);
end

--[[
	Helper to send a drilled into value to the client.
--]]
function Debugger:__sendVar(var)	
	self._lastAccessedVar = var

	local chainEnd = nil

	if type(var) ~= "object" then
		chainEnd = true
	end

	self:_sendAndEval({t="var", chainEnd=chainEnd, var=var })
end

function Debugger:_getUpVals(level)
	local vars = {}
	local func = debug.getinfo(level, "f").func
	do local i = 1; repeat
		local name, value = debug.getupvalue(func, i)
		if name then vars[name] = tostring(value) end
		i = i + 1
	until name == nil end

	return vars
end

function Debugger:_listUpVals(level)
	local vars = self:_getUpVals(level)

	self:_sendAndEval({t = "u", vars = vars})
end

function Debugger:_doEval(code)	
	local compile, func = pcall(loadstring, "return " .. code)

	if not compile then
		self:_sendAndEval({t = "eval", returnValue = func})
	end

	if self._breakFunc and func then
		print("setting function env")
		-- Set the environment of the code to the breakpoint 
		local mainEnv = getfenv(self._breakFunc)
		local level = self:_findFuncLevel(1)
		local localVars = self:_getLocals(level)
		local upValues = self:_getUpVals(level)
		setmetatable(upValues, {__index = mainEnv})
		setmetatable(localVars, {__index = upValues})
		setfenv(func, localVars)
	end

	local returnValues
	if func then
		returnValue = pack(pcall(func))
	end

	local list = {}
	for i,v in ipairs(returnValue) do
		list[#list + 1] = tostring(v)
	end

	self:_sendAndEval({t = "eval", returnValue = list})
end
