--------------------------------------------------------------------------------------------
--	EventDispatcher.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Provides event dispatching functionality through the use of event queues and callbacks.
--  To use the EventDispatcher in your code inherit from the dispatcher. Create an EVENTS
--  table that contains a property and short string per event type. For example:
--  ClassName.EVENTS = {
--		StatusChanged = "sc",
--		MessageReceived = "mc",
--	}
--  Use the addEventListener and removeEventListener methods to register callbacks for events
--  callbacks are called in the order they are registered.
--  A class that inherits from EventDisptcher can dispatch an event by calling dispatchEvent. 
--------------------------------------------------------------------------------------------

require("law.objects")
require("law.Log")

--------------------------------------------------------------------------------------------
--										EVENT CLASS										  --
--------------------------------------------------------------------------------------------

Event = inheritsFrom(nil)

--[[
	@param target The object the event came from.
	@param type The event type
	@param tParams The table of event parameters
--]]
function Event:init(target, type, tParams)
	self.target = target
	self.type = type
	self.params = tParams
end

--------------------------------------------------------------------------------------------
--									 END EVENT CLASS									  --
--------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------
--									EVENTDISPATCHER CLASS								  --
--------------------------------------------------------------------------------------------

EventDispatcher = inheritsFrom(nil)

EventDispatcher.eventTableName = "__eventDispatcher_eventQueue"

function EventDispatcher:init()
	Log:Trace("EventDispatcher:init()")
	self[self.eventTableName] = {}
end

function EventDispatcher:dispatchEvent(eventType, tParams)
	local queue = self:_getEventQueue(eventType)

	if not queue then
		return
	end

	local event = Event:create(self, eventType, tParams)

	for i, eventListener in ipairs(queue) do
		eventListener(event)
	end
end

function EventDispatcher:addEventListener(eventType, callbackFunction)
	Log:Debug("EventDispatcher:addEventListener", eventType, callbackFunction)
	if type(callbackFunction) ~= "function" then
		error("EventDispatcher:addEventListener callbackFunction is not of type 'function'")
		return
	end

	if not self:isValidEventType(eventType) then 
		error("EventDispatcher:addEventListener cannot insert event, invalid event type")
		return
	end

	local queue = self:_getEventQueue(eventType, true)

	table.insert(queue, callbackFunction)
end

function EventDispatcher:removeEventListener(eventType, callbackFunction)
	local queue = self:_getEventQueue()

	if queue then
		for i,v in ipairs(queue) do
			if v == callbackFunction then
				table.remove(queue, i)
				return
			end
		end
	end
end

--[[
	Clears the event queues.
--]]
function EventDispatcher:clear()
	self[self.eventTableName] = {}
end

function EventDispatcher:isValidEventType(eventType)
	Log:Debug("EventDispatcher:isValidEventType", eventType)
	Log:PrintTable(self.EVENTS)
	for k,v in pairs(self.EVENTS) do
		if v == eventType then
			return true
		end
	end
	return false
end

--[[
	Gets the event queue forthe eventType. If true is passed to createIfNone then
	the queue will be created if it doesn't exist
--]]
function EventDispatcher:_getEventQueue(eventType, createIfNone)
	local queue = self[self.eventTableName][eventType]

	if not queue and createIfNone then
		queue = {}
		self[self.eventTableName][eventType] = queue
	end

	return queue
end

function __test_EventDispatcher()
	local foo = inheritsFrom(EventDispatcher)
	
	foo.EVENTS = {
		event1 = "1",
		event2 = "2",
		all = "all"
	}

	local foo1 = foo:new()
	local foo2 = foo:new()
	local foo3 = foo:new()
	local foo4 = foo:new()

	local callback11 = function(evt)
		print("callback1")
		assert(evt.target == foo1, "Callback11 called with invalid event target")
		assert(evt.type == foo1.EVENTS.event1, "Callback11 called with invalid event type")
	end

	local callback21 = function(evt)
		print("callback2")
		assert(evt.target == foo2, "Callback21 called with invalid event target")
		assert(evt.type == foo2.EVENTS.event1, "Callback21 called with invalid event type")
	end

	local callback12 = function(evt)
		print("callback1")
		assert(evt.target == foo1, "Callback12 called with invalid event target")
		assert(evt.type == foo1.EVENTS.event2, "Callback12 called with invalid event type")
	end

	local callback22 = function(evt)
		print("callback2")
		assert(evt.target == foo2, "Callback22 called with invalid event target")
		assert(evt.type == foo2.EVENTS.event2, "Callback22 called with invalid event type")
	end

	local numAllFired = 0
	local callbackAll1 = function(evt)
		all1Fired = true
		numAllFired = numAllFired + 1
	end

	local callbackAll2 = function(evt)
		all2Fired = true
		numAllFired = numAllFired + 1
	end
	
	foo1:addEventListener(foo1.EVENTS.event1, callback11)
	foo2:addEventListener(foo2.EVENTS.event1, callback21)
	foo1:addEventListener(foo1.EVENTS.event2, callback12)
	foo2:addEventListener(foo2.EVENTS.event2, callback22)

	foo1:addEventListener(foo1.EVENTS.all, callbackAll1)
	foo1:addEventListener(foo1.EVENTS.all, callbackAll2)

	foo1:dispatchEvent(foo1.EVENTS.event1, nil)
	foo2:dispatchEvent(foo2.EVENTS.event1, nil)
	foo1:dispatchEvent(foo1.EVENTS.event2, nil)
	foo2:dispatchEvent(foo2.EVENTS.event2, nil)
	foo1:dispatchEvent(foo1.EVENTS.all, nil)

	print(numAllFired)
	assert(all1Fired, "All 1 callback did not fire")
	assert(all2Fired, "All 2 callback did not fire")
	assert(numAllFired == 2, "The number times the all event fired is incorrect.")
end
