--------------------------------------------------------------------------------------------
--	Lru.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Implements a least recently used cache in Lua
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.LinkedList"

Lru = inheritsFrom(nil)

--[[
	Creates an LRU that can have maxSize items. When maxSize is hit 
	purgeSize items will be removed. purgeSize must be less than maxSize

	@param maxSize The total number of items the cache can contain
	@param purgeSize The number of items to purge after the maxSize has been exceeded
	@param name The name of the cache. Used for logging and 
--]]
function Lru:init(maxSize, purgeSize, name)
	assert(maxSize > purgeSize, "Lru maxSize must be greater than purgeSize")
	self._name = name
	self._maxSize = maxSize
	self._purgeSize = purgeSize
	self._list = LinkedList:new()
	self._map = {}
end

function Lru:length()
	return self._list:length()
end

function Lru:log(...)
	if Log then 
		Log:Debug(...)
	else
		print(...)
	end
end

--[[
	Inserts an item into the cache.
--]]
function Lru:insert(key, value)
	if self._map[key] then
		return -- already in map
	end

	self._map[key] = self._list:add{key = key, value = value}

	if self._list:length() > self._maxSize then
		self:log("Purging cache: ", self._name)
		local numToPurge = self._purgeSize
		for item in self._list:all() do
			self._map[item.val.key] = nil
			self._list:remove(item)
			numToPurge = numToPurge - 1
			if numToPurge <= 0 then
				break
			end
		end
	end
end

--[[
	Finds an item in the cache. If the item is found then
	a "hit" is performed and the item is moved to the top (most recently used)
	of the list

	@return The item found or nil
--]]
function Lru:find(key)
	local linkedItem = self._map[key]

	if not linkedItem then
		return nil
	end
	
	-- Hit
	self._list:remove(linkedItem)
	self._map[key] = self._list:add(linkedItem.val)

	return linkedItem.val.value
end

function Lru:clear()
	self._list = LinkedList:new()
	self._map = {}
end

function Lru:__test_Lru()
	local lru = Lru:new(10, 5, "unit test cache")

	assert(lru, "No lru")
	
	for i=1,9 do
		lru:insert(i, i)
		assert(lru._map[i].val.key == i, "Map key values are not correct")
		assert(lru._list._end == lru._map[i], "List end is not equal to the last item inserted")
	end

	for i=1,9 do
		local val = lru:find(i)
		assert(val == i, "Value is not correct")
		assert(lru._list._end.val.key == i, "Found item wasn't moved to the end")
	end

	-- perform purge
	lru:insert(10,10)
	lru:insert(11,11)
	
	-- Count items in the map
	local count = 0
	for _ in pairs(lru._map) do
		count = count + 1
	end
	
	assert(count == 6, "Number of items in map table is not 6 it's " .. count)
	assert(lru._list:length() == 6, "Number of items in list is not 6 it's " .. lru._list:length())
	print("Test done")
end
