require "law.objects"
require "law.errors"

local function null_or_unpack(val)
	if val then
		return unpack(val, 1, table.maxn(val))
	else
		return nil
	end
end

Promise = inheritsFrom(nil)

Promise.States = {
	Pending = "pending",
	Rejected = "rejected",
	Resolved = "resolved",
}

Promise.Events = {
	Progress = "progress",
	Fail = "fail",
	Done = "done",
	Always = "always"
}

function Promise:init()
	self._state = self.States.Pending
	self.is_deferred = true
	self._callbacks = {}
end

function Promise:state()
	return self._state
end

function Promise:value()
	return null_or_unpack(self._value)
end


function Promise:reject(...)
	local arg = {...}
	assert(self:state() == Promise.States.Pending)
	self._value = arg
	self._state = Promise.States.Rejected

	for i,v in ipairs(self._callbacks) do 
		if v.event == Promise.Events.Always or v.event == Promise.Events.Fail then
			v.callback(...)
		end
	end
	self._callbacks = {}
end

function Promise:resolve(...)
	local arg = {...}
	assert(self:state() == Promise.States.Pending, "Promise:resolve() already resolved.\n" .. debug.traceback())
	self._value = arg
	self._state = Promise.States.Resolved

	for i,v in ipairs(self._callbacks) do
		if v.event == Promise.Events.Always or v.event == Promise.Events.Done then
			v.callback(...)
		end
	end
	self._callbacks = {}
end

function Promise:notify(...)
	assert(self:state() == Promise.States.Pending)

	for i,v in ipairs(self._callbacks) do
		if v.event == Promise.Events.Progress then
			v.callback(...)
		end
	end
end

function Promise:always(callback)
	if self:state() ~= Promise.States.Pending then
		callback(null_or_unpack(self._value))
	else
		table.insert(self._callbacks, { event = Promise.Events.Always, callback = callback })
	end
	return self
end

function Promise:done(callback)
	if self:state() == Promise.States.Resolved then
		callback(null_or_unpack(self._value))
	elseif self:state() == Promise.States.Pending then
		table.insert(self._callbacks, { event = Promise.Events.Done, callback = callback})
	end
	return self
end

function Promise:fail(callback)
	if self:state() == Promise.States.Rejected then
		callback(null_or_unpack(self._value))
	elseif self:state() == Promise.States.Pending then
		table.insert(self._callbacks, {event = Promise.Events.Fail, callback = callback})
	end
	return self
end

function Promise:progress(callback)
	if self:state() == Promise.States.Pending then
		table.insert(self._callbacks, {event = Promise.Events.Progress, callback = callback})
	end
	return self
end




function Promise:chain(onDone, onFail)
	local p = Promise:new()

	local function doneCallback(...)
		local retVals = {pcallAndPrint(onDone, ...)}

		if retVals[1] == false then
			p:reject(retVals[2])

		elseif type(retVals[2]) == "table" and
			retVals[2].isa and
			retVals[2]:isa(Error) then

			p:reject(retVals[2])

		elseif type(retVals[2]) == "table"  and 
			   retVals[2].isa and 
			   retVals[2]:isa(Promise) then

			retVals[2]:done(
				function(...)
					--print("p:resolve as promise")
					p:resolve(...)
				end
			)

			retVals[2]:fail(
				function(...)
					--print("p:reject as promise")
					p:reject(...)
				end
			)										
		else
			--print("p:resolve not a promise")
			p:resolve(select(2, null_or_unpack(retVals)))
		end					
	end

	local function failCallback(...) 					
		p:reject(onFail(...)) 
	end

	if onDone and self._state == Promise.States.Resolved then
	
		-- Call the callback
		doneCallback(null_or_unpack(self._value))
		return p

	elseif onDone then
		self._callbacks[#self._callbacks + 1] = 
		{
			event = Promise.Events.Done, 
			callback = doneCallback,
		}
	end

	if onFail and self._state == Promise.States.Rejected then
		-- Call the callback
		failCallback(null_or_unpack(self._value))
		return p
		
	elseif onFail then
		self._callbacks[#self._callbacks + 1] = 
		{
			event = Promise.Events.Fail, 
			callback = failCallback,
		}
	end

	return p
end

function Promise:finish(onDone, onFailed)
	if onDone then
		self:done(onDone)
	end

	if onFailed then
		self:fail(onFailed)
	end
end


local NIL = {} -- placeholder value for nil, storable in table.
local function packWithNil(...)
  local n = select('#', ...)
  local t = {...}
  for i = 1,n do
    if t[i] == nil then
      t[i] = NIL
    end
  end
  return t
end

local function unpackWithNil(t, k, n)
  k = k or 1
  n = n or #t
  if k > n then return end
  local v = t[k]
  if v == NIL then v = nil end
  return v, unpackWithNil(t, k + 1, n)
end

--[[
	Static method
--]]
function Promise.WhenWithPromises(...)
	local arg = packWithNil(...)
	local deferred = Promise:new()
	local returns = {}
	local promises = {}	
	local total = #arg
	local completed = 0
	local failed = 0

	local isCompleted = function()
		if completed == total and deferred:state() == Promise.States.Pending then

			-- We need to unpack all of the packed arguments and maintain order
			-- and nils.
			local linearList = {}

			for i=1, total do
				local packedArgs = returns[i]
				for i,v in ipairs(packedArgs) do
					linearList[#linearList + 1] = v
				end
			end

			--print("When returning:", #linearList, "items")

			-- Unpack all of the returns and keep them in order.
			returns = {unpackWithNil(linearList)}

			if failed > 0 then
				--print("When failed")
				--deferred:reject(null_or_unpack(returns))
				deferred:reject(returns, promises)
			else
				--deferred:resolve(null_or_unpack(returns))
				--print("Resolving deferred WHEN promise")
				deferred:resolve(returns, promises)
			end
		end
	end

	for i,v in ipairs(arg) do
		promises[i] = v
		if type(v) == "table" and 
			v.isa and 
			v:isa(Promise) and 
			v.is_deferred then
			--print("Setting up promise")
			local promise = v
			v:always(
				function(...)
					if promise:state() == 'rejected' then
						failed = failed + 1
					end
					completed = completed + 1

					-- There might be multiple return values
					-- we are going to save their order until all
					-- values are stored.
					returns[i] = packWithNil(...)
					--print("promise done")
					isCompleted()
				end
			)
		else
			-- Pack for promises that return multiple values.
			returns[i] = packWithNil(v)
			completed = completed + 1
		end
	end

	-- Check if all of the promises are done
	isCompleted()

	return deferred
end

--[[
	Allows waiting for a number of variables including promises. 
	@return A promise that will wait for all return values.
--]]
function Promise.When(...)
	local whenPromise = Promise:new()
	local p = Promise.WhenWithPromises(...)
	-- Strip the promises and unpack the returns
	p:done(
		function(returns, promises)
			--print("p done", #p._callbacks)
			whenPromise:resolve(null_or_unpack(returns))
		end
	)

	p:fail(
		function(returns, promises)
			--print("p fail")
			whenPromise:reject(null_or_unpack(returns))
		end
	)

	return whenPromise
end

function Promise.test()
	local function t1(msg)
		print "t1"
		return msg, "t1"
	end

	local function t2(msg, msg2)
		print "t2"
		return msg, msg2, "t2"
	end

	local function t3(msg, msg2, msg3)
		print "t3"
		return msg, msg2, msg3, "t3"
	end

	local function fail(msg)
		print "failed"		
	end

	local function done(...)
		print("recvd msgs:", ...)
	end

	local waitPromise = Promise:new()

	local function wait(msg, msg1)
		print("waiting:", msg, msg1)
		return waitPromise
	end

	local promise = Promise:new()

	promise
		:chain(t1, fail)
		:chain(t2, fail)
		:chain(t3, fail)
		:finish(done, fail)

	promise:resolve("Hello")

	promise = Promise:new()
	promise:chain(t1, 		fail)
		   :chain(t2, 		fail)
		   :chain(t3, 		fail)
		   :finish(done,  fail)

	promise:reject("failed")

	promise = Promise:new()
	promise
		:chain(t1, fail)
		:chain(wait, fail)
		:chain(t2, fail)
		:chain(t3, fail)
		:finish(done, fail)
	
	promise:resolve("wait for it")
	waitPromise:resolve("Done waiting", "wait")

	waitPromise = Promise:new()
	local whenPromise = Promise.When(waitPromise, "5", nil)
	whenPromise:done(
		function(...)
			print(...)
		end
	)
	waitPromise:resolve("We are done", "or so we think", "I hope this shows up")

	waitPromise = Promise:new()
	local whenPromise = Promise.When(waitPromise, "5", nil)
	whenPromise:done(
		function(...)
			print(...)
		end
	)
	waitPromise:resolve(nil, 1, nil, 2, nil)
end

