--------------------------------------------------------------------------------------------
--	RoomAudioQueue.lua
--
--	Represents an audio queue of data that the digital audio server will
--  play. Each queue is associated with at least one room.
--
--				   		          |Shuffle|Track |
--				 |Clear Q?|Goto NP| Queue |Order | Playhead
--	---------------------------------------------------------
--	Play 		 |    X   |   X   |   O   |1->N  |  1
--	Shuffle Play |    X   |   X   |   X   |Rand  |  1
--	Add To Queue |    O   |   O   |  N.C. |N.C.  |  N.C.
--	Play Next    |    O   |   O   |   O   |N.C.  |  N.C.
--
--  Copyright 2014 Control4 Corporation. All rights reserved.
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"
require "law.Xml"
require "law.StringUtils"
require "law.BaseAudioQueue"
require "law.ServiceLocator"

RoomAudioQueue = inheritsFrom(BaseAudioQueue)

RoomAudioQueue.AddFlags = {
	Play 		= "PLAY", -- To the beginning of the queue, Starts playing the item. Clears the queue
	PlayNext  	= "PLAY_NEXT", -- After the currently playing item. Turns off shuffle
	AddToQueue  = "ADD_TO_QUEUE", -- At the end of the queue
	ShufflePlay = "SHUFFLE_PLAY", -- Puts the queue in shuffle mode and plays. Clears the queue
}


function RoomAudioQueue:init(roomId, proxyBindingId)
	-- Initialize BaseAudioQueue
	self:super():init(roomId, proxyBindingId)

	Log:Debug("RoomAudioQueue:init, self:", tostring(self))

	-- The actual queue of items
	self._queue = {}
	-- Used when in shuffle mode. This queue stores the shuffled track order and is displayed
	-- when in shuffled mode.
	self._shuffleQueue = {}
	-- Set the currently playing index to 0 because nothing is playing.
	self._index = 0
	self._shuffleIndex = 0

	-- Initialize the modes
	local settings = ServiceLocator.AudioQueueManager:getQueueSettingsByRoom(roomId)
	self._shuffleMode = settings.shuffleVal
	self._repeatMode = settings.repeatVal

	self._updateQueueSettingsDel = function(evt)
		self:updateSettings(evt.params)
	end

	-- Register for settings changes
	ServiceLocator.AudioQueueManager:addEventListener(AudioQueueManager.EVENTS.RoomSettingsChanged, self._updateQueueSettingsDel)
end

--[[
	Called when the the queue is being deleted.
--]]
function RoomAudioQueue:destroy(roomOff)
	self:_clearQueue()
	self:queueChanged()

	ServiceLocator.AudioQueueManager:removeEventListener(AudioQueueManager.EVENTS.RoomSettingsChanged, self._updateQueueSettingsDel)

	self:super():destroy(roomOff)
end


--[[
	Adds an audio item to the zone queue. UIs are informed that
	the queue has changed when this method is called.
	@param item An array of AudioQueueItems
	@param flag One of the RoomAudioQueue.AddFlags types.
			    Indicates where in the queue the items should be added
--]]

function RoomAudioQueue:addAudioItems(items, flag, fromProgramming, volume, shuffleMode, repeatMode)
	local isNewQueue = (#self._queue == 0)
	if isNewQueue then
		self._volume = volume
		if shuffleMode then
			self._shuffleMode = (shuffleMode == "1")
		end
		if repeatMode then
			self._repeatMode = (repeatMode == "1")
		end
	end

	-- honor previous settings if from programming.
	if fromProgramming and flag == self.AddFlags.Play and self._shuffleMode then
		flag = self.AddFlags.ShufflePlay
	end

	-- Should the queue be cleared?
	if flag == self.AddFlags.Play or
	   flag == self.AddFlags.ShufflePlay then
	   self:_clearQueue()
	end

	local insertionPoint
	local shuffleInsertionPoint

	if flag == self.AddFlags.Play then
	   self._shuffleMode = false
	   insertionPoint = 1

	elseif flag == self.AddFlags.PlayNext then
		self._shuffleMode = false;
		self._shuffleQueue = {}
		self._shuffleIndex = 0
		insertionPoint = self._index + 1

	elseif flag == self.AddFlags.AddToQueue then
		insertionPoint = #self._queue + 1
		if self._shuffleMode then
			shuffleInsertionPoint = #self._shuffleQueue + 1
		end

	elseif flag == self.AddFlags.ShufflePlay then
		self._shuffleMode = true
		insertionPoint = 1
		shuffleInsertionPoint = 1
	end

	-- Do the insertion
	local tempIndex = insertionPoint
	for i,v in ipairs(items) do
		table.insert(self._queue, tempIndex, v)
		tempIndex = tempIndex + 1
	end

	if self._shuffleMode then
		local shuffledItems = self:_randomize(items)
		tempIndex = shuffleInsertionPoint
		for i,v in ipairs(shuffledItems) do
			table.insert(self._shuffleQueue, tempIndex, v)
			tempIndex = tempIndex + 1
		end
	end

	if flag == self.AddFlags.Play or
	   flag == self.AddFlags.ShufflePlay or
	   isNewQueue then
	   self:playIndex(1)
	elseif self._state == self.States.Playing then
		-- Update the information about the next track
		-- This could be optimized to only be done if it needs to be done
		self:_setNextItem(self:_getNextItem())
	end

	self:queueChanged()
end

--[[
	Won't dispatch a QueueChanged event.
--]]
function RoomAudioQueue:_clearQueue()
	Log:Debug("RoomAudioQueue:_clearQueue()")
	self._queue = {}
	self._shuffleQueue = {}
	self._index = 0
	self._shuffleIndex = 0
end

--[[
	Randomizes an array of items returning a new array
	@param items A list of items in their original order.
	@param firstIndex the index of the items in "items"
			that should be the first item in the randomized
			array. This is used to keep the first item of
			the array at the top.
	@return Array a new array of items in a randomized order
--]]
function RoomAudioQueue:_randomize(items, firstIndex)
	-- Copy the original array.
	local remainingItems = tableConcat({}, items)
	local randomizedItems = {}

	if firstIndex then
		local item = table.remove(remainingItems,firstIndex)
		randomizedItems[1] = item
	end

	while #remainingItems > 0 do
		local index = math.random(#remainingItems)
		local item = remainingItems[index]
		table.remove(remainingItems, index)
		randomizedItems[#randomizedItems + 1] = item
	end

	return randomizedItems
end

--[[
	Removes and audio item by id from the queue. UIs are informed
	that the queue has changed when this method is called.
--]]
function RoomAudioQueue:removeAudioItemById(id)
	Log:Debug("RoomAudioQueue:removeAudioItemById() Removing id from queue:", id)

	-- Find the item in the queue
	local index = self:_findIndexById(id, self._queue)

	if index and index > 0 then
		table.remove(self._queue, index)
		if index < self._index then
			self._index = self._index - 1
		end
	end

	local shuffleIndex
	if self._shuffleMode then
		shuffleIndex = self:_findIndexById(id, self._shuffleQueue)

		if shuffleIndex and shuffleIndex > 0 then
			table.remove(self._shuffleQueue, shuffleIndex)
			-- Adjust currently playing index up one.
			if shuffleIndex < self._shuffleIndex then
				self._shuffleIndex = self._shuffleIndex - 1
			end
		end
	end

	-- Check if the currently playing track was removed.
	if #self._queue == 0 then
		self:clearQueue()

	elseif self._shuffleMode and shuffleIndex and
		   self._shuffleIndex == shuffleIndex then
		   	Log:Debug("Removed currently playing track. Playing next shuffle index:", shuffleIndex)
			self:playIndex(shuffleIndex)

	elseif index and self._index == index then
			Log:Debug("Removed currently playing track. Playing next index:", index)
			self:playIndex(index)
	elseif self._state == self.States.Playing then
		-- Update the information about the next track
		-- This could be optimized to only be done if it needs to be done
		self:_setNextItem(self:_getNextItem())
	end

	self:queueChanged()
end

--[[
	Moves the audo item with the id to the next item
	to be played in the queue.
--]]
function RoomAudioQueue:moveToNextById(id)
	Log:Debug("RoomAudioQueue:moveToNextById(" .. id .. ")")

	-- Find the item in the queue
	local index = self:_findIndexById(id, self._queue)
	local audioItem = self._queue[index]


	-- Check if we are trying to playNext the currently playing track.
	-- If so don't do anything.
	if index == self._index then
		return
	end


	if self._shuffleMode then
		local shuffleIndex = self:_findIndexById(id, self._shuffleQueue)

		if self._shuffleIndex > shuffleIndex then
			self._shuffleIndex = self._shuffleIndex - 1
		end

		local nextIndex = self._shuffleIndex + 1

		if shuffleIndex and shuffleIndex > 0 then
			table.remove(self._shuffleQueue, shuffleIndex)
			table.insert(self._shuffleQueue, nextIndex, audioItem)
		end
	else
		-- Adjust the currently playing index because
		-- we removed an item above it.
		if self._index > index then
			self._index = self._index - 1
		end

		local nextIndex = self._index + 1

		if index and index > 0 and
		   nextIndex and nextIndex > 0 then
			table.remove(self._queue, index)
			table.insert(self._queue, nextIndex, audioItem)
		end
	end

	-- Update the information about the next track
	self:_setNextItem(self:_getNextItem())

	self:queueChanged()
end

--[[
	Shuffles the items in the queue. Notifies UIs of a queue change.
--]]
function RoomAudioQueue:toggleShuffle()
	Log:Debug("Toggle shuffle mode to: ", not self._shuffleMode)

	self._shuffleMode = not self._shuffleMode

	-- Conversion for digital audio.
	local numVal = 0
	if self._shuffleMode then
		numVal = 1
	end

	-- Notify digital audio of the change.
	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "SHUFFLE", {ROOM_ID = self:getOwningRoom(), SHUFFLE = numVal})

	-- Check if this is a transition to off
	if self._shuffleMode == false then
		self._shuffleQueue = {}
		self._shuffleIndex = 0

		if self._state == self.States.Playing then
			-- Update or clear the information about the next track
			self:_setNextItem(self:_getNextItem())
		end

		self:queueChanged()
		return
	end

	-- Randomize the current queue.
	-- Keep the currently playing item at the top.
	self._shuffleQueue = self:_randomize(self._queue, self._index)

	-- Find the currently play song
	self._shuffleIndex = 0

	local id
	if self._index >= 1 then
		id = self._queue[self._index]:getId()
	end

	-- Keep the now playing index the same
	if id then
		self._shuffleIndex = self:_findIndexById(id, self._shuffleQueue)

		if self._state == self.States.Playing then
			-- Update or clear the information about the next track
			self:_setNextItem(self:_getNextItem())
		end
	elseif self._state == self.States.Playing then
		-- Clear the information about the next track
		self:_setNextItem(nil)
	end

	-- Notify navigators of the change
	self:queueChanged()
end

function RoomAudioQueue:toggleRepeat()
	self._repeatMode = not self._repeatMode

	-- Conversion for digital audio
	local numVal = 0
	if self._repeatMode then
		numVal = 1
	end

	-- Notify digital audio of the change.
	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "REPEAT", {ROOM_ID = self:getOwningRoom(), REPEAT = numVal})

	if self._state == self.States.Playing then
		-- Update or clear the information about the next track
		-- Could be optimized to only do it if the currently playing item is the last
		self:_setNextItem(self:_getNextItem())
	end

	self:dashboardChanged ()
end

--[[
	Finds the index of an item by its id in the given queue.
	@param id The id of the item to find
	@param queue The queue to search
	@return The index of the item or 0 if not found
--]]
function RoomAudioQueue:_findIndexById(id, queue)
	for i, item in ipairs(queue) do
		if item:getId() == id then
			return i
		end
	end

	return 0
end

--[[
	Starts playing the queue from a stopped or paused state. This might
	notify the UI that the now playing has changed.
	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:play()
	Log:Debug("RoomAudioQueue:play()")
	if self._state == self.States.Playing then
		Log:Debug("RoomAudioQueue:play() Play -> Play index:", self._index)
		return -- Nothing to do

	elseif self._state == self.States.Paused then
		Log:Debug("RoomAudioQueue:play() Pause -> Play index:", self._index)
		self._state = self.States.Playing
		self:dashboardChanged()
		return

	elseif self._state == self.States.Stopped then
		if #self._queue > 0 then
			Log:Debug("RoomAudioQueue:play() Stop -> Play index: 1")
			self:playIndex(1) -- this will change the state.
			return
		else
			Log:Error("RoomAudioQueue:play() Nothing in queue to play")
		end
	elseif self._state == self.States.Empty then
		if #self._queue > 0 then
			Log:Debug("RoomAudioQueue:play() Empty -> Play index: 1")
			self:playIndex(1)
		else
			Log:Error("RoomAudioQueue:play() Trying to play an empty queue")
		end
	else
		Log:Error("RoomAudioQueue:play() Unknown state")
	end
end

--[[
	Puts the queue in the stopped state. This usually comes from the
	room. If it doesn't come from the room then we should tell the room.
--]]
function RoomAudioQueue:stop(tellRoom)
	if not tellRoom then
		self:super():stop()
	else
		C4:SendToDevice(self:getOwningRoom(), "STOP", {})
	end
end

--[[
	Called when the end of the currently playing item has been reached.
	@param queueInfo Can be nil, is usually a timestamp.
--]]
function RoomAudioQueue:endOfItem(queueInfo)
	self:super():endOfItem(queueInfo)

	local queue = self:_getCurrentQueue()
	local index = self:getNowPlayingIndex()

	-- Check the state of the queue to make sure we should keep playing.
	-- This is to prevent stopping the queue and then getting a queue end
	-- and going on and playing
	if self._state == self.States.Stopped then
		return
	end

	if index ~= #queue then
		self:playIndex(index + 1)
	elseif self._repeatMode then
		self:playIndex(1)
	else
		-- We are done playing the queue. Stop it and wait for delete
		self:stop()

		-- There is a bug that when the last song ends it repeats...tell digital audio to stop playing.
		-- for i, roomId in ipairs(self._rooms) do
		-- 	C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "STOP", {ROOM_ID = roomId});
		-- end
		C4:SendToDevice(AudioQueueManager.DigitalAudioDeviceId, "CLEAR_QUEUE", {ROOM_ID=self:getOwningRoom()})
	end
end

--[[
	Gets the current audio queue depending on the mode of operation
--]]
function RoomAudioQueue:_getCurrentQueue()
	if self._shuffleMode then
		return self._shuffleQueue
	end

	return self._queue
end

--[[
	Gets the queue of items
--]]
function RoomAudioQueue:getQueueItems()
	return self:_getCurrentQueue()
end

--[[
	Plays an item identified by the items id.
	@param id The queues id. Not a trackId.

	@note trackId is not used becuase multiple items in the
	queue could have the same track id.

	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:playItemById(id)
	local currentQueue = self:_getCurrentQueue()

	local index = self:_findIndexById(id, currentQueue)
	if index > 0 then
		self:playIndex(index)
		return
	end

	Log:Warn("RoomAudioQueue:playItemById, Could not find item with id:", id)
end

--[[
	Plays an item at the specified index
--]]
function RoomAudioQueue:playIndex(index, isNext)
	if index < 1 or index > #self._queue then
		Log:Error("RoomAudioQueue:playIndex() index is out of range")
		return
	end

	self._state = self.States.Playing
	-- Update the new now playing index
	self:_setPlayIndexes(index)

	-- It doesn't matter which index we use to play the item
	-- they should both point to the same item in their respective queues.
	local item = self._queue[self._index]
	if isNext then
		self:_updateMediaInfo(item)
		self:_setNextItem(self:_getNextItem())
	else
		self:_playItem(item, self:_getNextItem())
	end
	self:dashboardChanged()
	self:nowPlayingChanged()
end

function RoomAudioQueue:_isNextItem(index)
	local queue = self:_getCurrentQueue()
	if index >= #queue and self._repeatMode then
		index = 1
	end
	return self:_getNextItemIndex() == index
end

function RoomAudioQueue:_getNextItemIndex()
	local queue = self:_getCurrentQueue() or {}
	local index = self:getNowPlayingIndex()
	if index < #queue then
		return index + 1
	elseif self._repeatMode then
		return 1
	end
	return 0
end

function RoomAudioQueue:_getNextItem()
	local queue = self:_getCurrentQueue()
	local index = self:_getNextItemIndex()
	if index > 0 then
		return queue[index]
	end
	return nil
end

function RoomAudioQueue:_setPlayIndexes(index)
	if self._shuffleMode then
		self._shuffleIndex = index
		self._index = self:_findIndexById(self._shuffleQueue[index]:getId(), self._queue)
	else
		self._index = index
	end
end

--[[
	Supplies information about the following track
--]]
function RoomAudioQueue:needNext(queueInfo)
	Log:Debug("RoomAudioQueue:needNext()")
	-- We were notified that the current track finished.
	-- Set the currently playing to the next track
	local playingNowIndex = self:_getNextItemIndex()

	if playingNowIndex > 0 then
		self:playIndex(playingNowIndex, true)
	else
		Log:Debug("Reached the end of the queue")
	end
end

--[[
	Returns the index of the currently playing audio item
--]]
function RoomAudioQueue:getNowPlayingIndex()
	if self._shuffleMode then
		return self._shuffleIndex
	else
		return self._index
	end
end

--[[
	Sends the QueueChanged event.

	@param navId [Optional] Send the event to a specific navigator.
			If left off the event is sent to all of the rooms that belong
			to the zone.

	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:queueChanged(navId)
	-- Encode the queue
	local index = self:getNowPlayingIndex()

	local tParams =
	{
		NowPlayingIndex = index, -- Normally we would subtract one but this queue has a header.
		List = self:queueToXml(),
		NowPlaying = Xml.BuildSimpleXml("NowPlaying", {shuffle = self._shuffleMode, ["repeat"] = self._repeatMode, actionIds="ShuffleQueue RepeatQueue SaveAsPlaylist"}, true),
	}

	self:_sendEvent("QueueChanged", tParams, navId, {"List", "NowPlaying"})
	if (USE_CUSTOM_DASH) then
		self:dashboardChanged()
	end
end

--[[
	Sends the currently playing items id to navigators.
	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:nowPlayingChanged()

	local tParams = {
		NowPlayingIndex = self:getNowPlayingIndex(), -- Normally we would subtract one from the index, but this queue has a header.
	}

	self:_sendEvent("QueueChanged", tParams)
end

--[[
	Sends the NowPlaying data in a QueueChanged event to the navigators.
--]]
function RoomAudioQueue:globalDataChanged()
	local tParams = {
		NowPlaying = Xml.BuildSimpleXml("NowPlaying", {shuffle = self._shuffleMode, ["repeat"] = self._repeatMode, actionIds="ShuffleQueue RepeatQueue SaveAsPlaylist"}, true)
	}

	self:_sendEvent("QueueChanged", tParams, nil, {"NowPlaying"})
	if (USE_CUSTOM_DASH) then
		self:dashboardChanged()
	end
end

--[[
	Converts the queue to the xml List format without the
	<List> container

	@param tData The data to convert from an array to xml.
	@param rawItemNames The name of table properties that are raw items.
			These properties are not converted to xml but are injected directly
			into the xml stream.
--]]
function RoomAudioQueue:queueToXml()
	if #self._queue == 0 then
		return "<List></List>"
	end

	local sumSeconds = 0
	local allowedActions = "PlayQueueItem PlayQueueTrackNext RemoveFromQueue ViewArtist ViewAlbum"

	if #self._queue == 1 then
		-- Don't allow a remove queue if there is only one item in the list. The queue state gets messed up.
		allowedActions = "PlayQueueItem PlayQueueTrackNext ViewArtist ViewAlbum"
	end

	local xml = StringBuilder:new()

	xml = xml .. "<List>"

	for i,v in pairs(self:_getCurrentQueue() or {}) do
		xml = xml .. v:toXml(allowedActions)
		sumSeconds = sumSeconds + v:getPlaybackSeconds()
	end

	xml = xml .. "</List>"

	local strSeconds = ""
	if sumSeconds > 0 then
		strSeconds = " - " .. StringUtils.secondsToTime(sumSeconds)
	end

	-- Add a header to the output
	local title

	if #self._queue == 1 then
		-- Singular case for localization
		title = string.make_localizable(gettext("{numberOfTracks} Track {numSeconds}"), {numberOfTracks = tostring(#self._queue), numSeconds = strSeconds})
	else
		-- Plural case for localization
		title = string.make_localizable(gettext("{numberOfTracks} Tracks {numSeconds}"), {numberOfTracks = tostring(#self._queue), numSeconds = strSeconds})
	end



	-- Add the header text
	xml:addAt(2, "<item><header>true</header><title>" .. title .. "</title><translate>true</translate></item>" )

	return tostring(xml)
end

--[[
	Called when the queue's settings have changed through programming.
--]]
function RoomAudioQueue:updateSettings(settings)
	Log:Debug("RoomAudioQueue:updateSettings()")
	Log:Trace("RoomAudioQueue:updateSettings() Settings >>>>>")
	Log:Trace(settings)
	Log:Trace("<<<<<")

	if settings.roomId ~= self:getOwningRoom() then
		return
	end

	if self._shuffleMode ~= settings.shuffleVal then
		self:toggleShuffle()
	end

	if self._repeatMode ~= settings.repeatVal then
		self:toggleRepeat()
	end
end

--------------------------------------------------------------------------------------------
--								  RECEIVED FROM PROXY METHODS						  	  --
--------------------------------------------------------------------------------------------

--[[
	Called when a SKIP_FWD is received
	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:SKIP_FWD(tParams)
	Log:Debug("RoomAudioQueue:SKIP_FWD")
	local queue = self:_getCurrentQueue()
	local index = self:getNowPlayingIndex()

	if index < #queue then
		self:playIndex(index + 1)
	elseif self._repeatMode then
		self:playIndex(1)
	else
		Log:Debug("Can't skip fwd. At the end of the queue")
	end
end

--[[
	Called when a SKIP_REV is received
	@overrides BaseAudioQueue
--]]
function RoomAudioQueue:SKIP_REV(tParams)
	Log:Debug("RoomAudioQueue:SKIP_REV")
	local index = self:getNowPlayingIndex()

	if self:secondsSinceLastPlay() > 15 then
		-- I want to hear it again...please.
		self:playIndex(index)

	elseif index ~= 1 then
		self:playIndex(index - 1)
	else
		Log:Debug("Can't skip back. At the beginning of the queue")
	end
end
--------------------------------------------------------------------------------------------
--								  RECEIVED FROM PROXY METHODS						  	  --
--------------------------------------------------------------------------------------------
