--------------------------------------------------------------------------------------------
--	StreamAudioQueue.lua
--	
--	Represents an audio queue that plays streams (stations). 
--
--  Copyright 2014 Control4 Corporation. All rights reserved.
--------------------------------------------------------------------------------------------
require "law.objects"
require "law.ServiceLocator"
require "law.BaseAudioQueue"
require "law.RoomAudioQueue"

StreamAudioQueue = inheritsFrom(BaseAudioQueue)

-- The interval to call the callback on in milliseconds.
StreamAudioQueue.CALLBACK_TIMEOUT = 10000

-- Only display the last x number of history items
StreamAudioQueue.HISTORY_MAX_LENGTH = 25

function StreamAudioQueue:init(roomId, proxyBindingId)
	self:super():init(roomId, proxyBindingId)
	Log:Debug("StreamAudioQueue:init() self:", tostring(self))

	-- Callback function to call on a timer
	self._callback = nil

	-- Holds the currently playing streams info
	self._streamItem = nil

	-- Holds the currently play track info
	self._currentTrack = nil

	-- List of stations history
	self._history = {}

	-- The callback timer
	self._timerPromise = nil
end	

function StreamAudioQueue:destroy(roomOff)
	if self._timerPromise then
		self._timerPromise:killTimer()
		self._timerPromise = nil
	end

	if self._callback then
		-- Tell the callback the queue is being destroyed.
		self._callback(self, nil, true)
	end

	self._callback = nil
	self._streamItem = nil
	self._currentTrack = nil
	self._history = nil

	self:super():destroy(roomOff)
end

function StreamAudioQueue:getHistory()
	return self._history
end

function StreamAudioQueue:getQueueItems()
	return self._history
end

--[[
	Starts the stream playing
	@streamItem An AudioQueueItem that describes the stream.	
	@dataCallback A function with the signature: function callback(thisQueue, AudioQueueItem, cleanup):AudioQueueItem, that
		will be called on an interval to check if the currently playing audio item has changed. This should
		return nil if it has not changed or an AudioQueueItem if it has. The first time the callback is called 
		the AudioQueueItem will be nil. Wrap the callback in a delegate if you need the function's self table otherwise
		no self will be provided. cleanup indicates the callback should cleanup and won't be used any longer by the queue. This occurs
		if the queue itself is destroyed or if startStream is call more than 1 time on a queue.
--]]
function StreamAudioQueue:startStream(streamItem, dataCallback)
	-- Cleanup the other callback
	if self._callback then
		self._callback(self, self._currentTrack, true)
	end
	
	self._streamItem = streamItem
	self._callback = dataCallback
	self._history = {}

	self:_playItem(streamItem, nil)
	self:queueChanged()
end

--[[
	This is called to set the currently playing item. The dataCallback
	from startStream() should call this method and pass in the currently playing audio
	item
--]]
function StreamAudioQueue:setCurrentItem(audioItem)
	if not self._currentTrack or 
	   audioItem:getUrl() ~= self._currentTrack:getUrl() then

	   self._currentTrack = audioItem
	   self:_setRoomInfo(audioItem)
	   table.insert(self._history, 1, audioItem)
	   if #self._history > self.HISTORY_MAX_LENGTH then
	   		table.remove(self._history)
	   end
	   self:queueChanged()
	end
end

--[[
	Sets the now playing info for all of the rooms participating in the zone.
--]]
function StreamAudioQueue:_setRoomInfo(item)
	for i,roomId in ipairs(self._rooms) do
	  local params = {}
	  params["ROOMID"] = roomId
	  params["IMAGEURL"] = item:findLargestImageUrl()

	  params["ALBUM"] = item:getAlbum()
	  params["ARTIST"] = item:getArtist()
	  params["TITLE"] = item:getTitle()
	  params["GENRE"] = ""
	  params["MEDIATYPE"] = "secondary"
	  --params["QUEUEID"] = queueId
	  params["MERGE"] = "True"
	  C4:SendToProxy(self._proxyBindingId, "UPDATE_MEDIA_INFO", params, "COMMAND")
  end
end

--[[
	
--]]
function StreamAudioQueue:_setupCallback()
	self._timerPromise = ServiceLocator.TimerManager:addTimer(self.CALLBACK_TIMEOUT)

	self._timerPromise:done(
		function()
			self._timerPromise = nil
			local track = self._callback(self, self._currentTrack)
			if track then
				self._currentTrack = track
				table.insert(self._history, 1, track)
				if #self._history > self.HISTORY_MAX_LENGTH then
					table.remove(self._history)
				end
				self:queueChanged()
			end
			--self:_setupCallback()
		end
	)
end

--[[
	@overrides BaseAudioQueue
--]]
function StreamAudioQueue:queueChanged()

	 -- Fix for race condition where the streamItem hasn't been set yet.
  	if not self._streamItem then
    	return
  	end

	local tParams = 
	{
		NowPlayingIndex = 1, -- We only ever play the station.
		List = self:queueToXml(),
		NowPlaying = string.format("<NowPlaying><actionIds>AddToLibrary SaveAsPlaylist</actionIds><stationId>%s</stationId></NowPlaying>", self._streamItem:getTrackId()),
	}

	self:_sendEvent("QueueChanged", tParams, navId, {"List", "NowPlaying"})
end

function StreamAudioQueue:queueToXml()

	local sb = StringBuilder:new()

	-- Start list
	sb:add("<List>")

		-- Radio station header
		sb:add("<item><title>")
		sb:add(gettext("Radio Station"))
		sb:add("</title><header>true</header></item>")

	if self._streamItem then 
		-- Add the station info
		sb:add(self._streamItem:toXml())

		if #self._history > 0 then
			-- History header
			sb:add("<item><title>")
			sb:add(gettext("Station History"))
			sb:add("</title><header>true</header></item>")

			-- History Items
			for i, item in ipairs(self._history or {}) do
				sb:add(item:toXml("AddTrackToLibrary ViewArtist ViewAlbum"))
			end
		end
	else
		sb:add("<item><title>")
		sb:add(gettext("Waiting for data..."))	
		sb:add("</title></item>")	
	end

	sb:add("</List>")
	return tostring(sb)
end

--[[	
	Convert the queue into a RoomAudioQueue 
--]]
function StreamAudioQueue:playItemById(id)
	Log:Debug("StreamAudioQueue:playItemById() converting queue to a RoomAudioQueue")
	-- Convert the stream to the a RoomAudioQueue and then start playing the item.
	local queue = ServiceLocator.AudioQueueManager:getQueueForRoom(self._rooms[1], true, RoomAudioQueue)
	queue:playItemById(id)
end

function StreamAudioQueue:play()
	Log:Debug("StreamAudioQueue:play()")

	if self._callback then
		self:_setupCallback()
	end

	if self._state == self.States.Playing then
		Log:Debug("StreamAudioQueue:play() Play -> Play index:", self._index)		
		return -- Nothing to do

	elseif self._state == self.States.Paused then
		Log:Debug("StreamAudioQueue:play() Pause -> Play index:", self._index)		
		self._state = self.States.Playing
		self:dashboardChanged()
		return	
	elseif self._state == self.States.Stopped then
		self._state = self.States.Playing
		self:dashboardChanged()
	else
		Log:Error("StreamAudioQueue:play() Unknown state")
	end
end

--[[
	Called when the stream is paused
	@override BaseAudioQueue
--]]
function StreamAudioQueue:pause()
	-- Stop the check timer while the stream is paused.
	if self._timerPromise then
		self._timerPromise:killTimer()
		self._timerPromise = nil
	end

	-- Notify super of pause
	self._state = self.States.Stopped
	self:dashboardChanged()
end

function StreamAudioQueue:stop()
	if self._timerPromise then
		self._timerPromise:killTimer()
		self._timerPromise = nil
	end

	self:super():stop()
end

function StreamAudioQueue:endOfItem(queueInfo)
	Log:Debug("StreamAudioQueue:endOfItem()")
	self:super():endOfItem(queueInfo)

	self:_playItem(self._streamItem, nil)
end

function StreamAudioQueue:nowPlayingChanged()
	Log:Debug("StreamAudioQueue:nowPlayingChanged")
end

--[[
	@overrides BaseAudioQueue
--]]
function StreamAudioQueue:dashboardChanged(navId)
	local buttons
	
	if self._state == self.States.Playing then
		buttons = {"Pause", "SkipFwd"}
	elseif self._state == self.States.Paused then
		buttons = {"Play"}
	elseif self._state == self.States.Stopped then
		buttons = {"Play"}
	elseif self._state == self.States.Deleted then
		buttons = {}
	else
		Log:Error("StreamAudioQueue:_dashboardChanged() Unknown state")
		return
	end

	local tParams = {
		Items = table.concat(buttons, " ")
	}

	-- The Room ids will be added in _sendEvent. That's how navigator's know 
	-- which rooms the dashboard applies to.
	self:_sendEvent("DashboardChanged", tParams, navId)
end

function StreamAudioQueue:AUDIO_URL_SELECTED(tParams)
	Log:Debug("StreamAudioQueue:AUDIO_URL_SELECTED")
	
	-- Update the dashboard
	self:play()

	if not self._timerPromise then
		self:_setupCallback()
	end
end

--[[
	Called when a SKIP_FWD is received
	@overrides BaseAudioQueue
--]]
function StreamAudioQueue:SKIP_FWD(tParams)
	Log:Debug("StreamAudioQueue:SKIP_FWD")
	self:_playItem(self._streamItem, nil)
	self:queueChanged()
end
