--------------------------------------------------------------------------------------------
--	StringBuilder.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Efficiently builds strings.
--
--  Usage:
--  Use the new method to create a StringBuilder object.
--  local builder = StringBuilder:new()
--
--  Use the add method to add a new string to the string builder
--  builder:add("string")
--
--  Use the tostring operator to get out the final string
--  print(builder) 
--  or
--  builder.tostring()
--------------------------------------------------------------------------------------------

require("law.objects")

StringBuilder = inheritsFrom(nil)



function StringBuilder:init()
	self._items = {}
	local mt = getmetatable(self)
	mt.__tostring = self.__tostring
	mt.__concat = self.__concat
end

function StringBuilder:add(right)
	--table.insert(self._items, right)
	self._items[#self._items + 1] = right
	return self
end

--[[
	Adds an item at an index in the builder strings table. Everything at or
	after the index is shifted right

	@param index The index to put the string.
	@param str The string that will be added
	@return self This object instance
--]]
function StringBuilder:addAt(index, str)
	table.insert(self._items, index, str)
	return self
end

--[[
	Joins the string with the given seperator. If no seperator
	is given the default is an empty string "".

	@param sep the Seperator to use. If nil then "" will be used.
	@return Returns the final string
--]]
function StringBuilder:join(sep)
	local seperator = sep
	if not seperator then
		seperator = ""
	end
	
	return table.concat(self._items, seperator)
end

function StringBuilder:__tostring()
	return table.concat(self._items)
end

function StringBuilder:__concat(str)
	self:add(str)
	return self
end



function __test_StringBuilder()
	local str = ""
	local start = os.clock()
	for i=1, 100000 do 
		str = str .. tostring(i)
	end
	print(str)
	local endTime = os.clock()

	print("ellapsed time normal concat:", endTime - start, "seconds")

	local builder = StringBuilder:new()
	start = os.clock()
	for i=1, 100000 do
		builder:add(tostring(i))
	end
	print(builder)
	endTime = os.clock()
	print("ellapsed builder time:", endTime - start, "seconds")
end
