--------------------------------------------------------------------------------------------
--  StringUtils.lua
--
--  Copyright 2015 Control4 Corporation. All rights reserved.
--  
--  Various string helpers. 
-- 
--------------------------------------------------------------------------------------------

--[[
    Splits a string into a table. Each item in the string is separated delim.

    @param str The string to split
    @param delim Astring delimeter
    @param maxNb [Optional] How many parts to split. 0 or nil is unlimited.
--]]
function string.split(str, delim, maxNb)
    -- Eliminate bad cases...
    if string.find(str, delim) == nil then
        return { str }
    end

    if maxNb == nil or maxNb < 1 then
        maxNb = 0    -- No limit
    end
    local result = {}
    local pat = "(.-)" .. delim .. "()"
    local nb = 0
    local lastPos
    for part, pos in string.gfind(str, pat) do
        nb = nb + 1
        result[nb] = part
        lastPos = pos
        if nb == maxNb then break end
    end
    -- Handle the last field
    if nb ~= maxNb then
        result[nb + 1] = string.sub(str, lastPos)
    end
    return result
end

--[[
    Creates a string that is driver localizable with variables.
    @param str The localizable string. Variable placeholders are of the format: {varName}
    @param vars A table of variable values. The table keys should match the variables that are in the str param.
--]]
function string.make_localizable(str, vars)
    -- Escape the translatable string
    local ret = "#!\"" .. str:gsub("\\", "\\\\"):gsub("\"", "\\\"") .. "\""
    if vars then
        -- Add on the variables and escape their values
        for var, value in pairs(vars) do
            ret = ret .. ";" .. var .. "=\"" .. tostring(value):gsub("\\", "\\\\"):gsub("\"", "\\\"") .. "\""
        end
    end
    return ret
end

--[[
    Trims a string
--]]
function string.trim(s)
    return s:match "^%s*(.-)%s*$"
end

--[[
    Trims leading spaces
--]]
function string.lTrim(s)
    return s:match "^%s*(.*)"
end

--[[
    Trims trailing spaces
--]]
function string.rTrim(s)
    return s:match "(.-)%s*$"
end

StringUtils = {}
--[[
    Converts the number of seconds into a time string.
--]]
function StringUtils.secondsToTime(seconds)
    local hours = math.floor(seconds / 3600)
    local mins = math.floor((seconds - hours * 3600) / 60)
    local remainSecs = seconds % 60

    if hours > 0 then
        return string.format("%d:%02d:%02d", hours, mins, remainSecs)
    else
        return string.format("%d:%02d", mins, remainSecs)
    end
end