--------------------------------------------------------------------------------------------
--	SystemEventManager.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--------------------------------------------------------------------------------------------

require("law.objects")
require("law.ServiceLocator")

function OnSystemEvent(data)
	if Log then 
		Log:Debug("System Event occurred - data: " .. data)
	else
		print("System Event ocurred - data:", data)
	end

	ServiceLocator.SystemEventManager:parseSystemEvent(data)
end

SystemEventManager = inheritsFrom(nil)

SystemEventManager._callbacks = {}

function SystemEventManager:init(addToServiceLocator)
	if addToServiceLocator then
		ServiceLocator:addService("SystemEventManager", self)
	end
end

--[[
	Registers for the dataToUI event
--]]
function SystemEventManager:registerDataToUI(driverId, callback)	
	local eventId = tonumber(C4SystemEvents["OnDataToUI"])

	self:registerEvent(eventId, driverId, callback)
end

--[[
	Unregisters previous events.
--]]
function SystemEventManager:unregisterDataToUI(driverId, callback)
	local eventId = tonumber(C4SystemEvents["OnDataToUI"])

	self:unregisterEvent(eventId, driverId, callback)
end

--[[
	Registers a system event to the callback
--]]
function SystemEventManager:registerEvent(eventId, driverId, callback)
	print("SystemEventManager:registerEvent", eventId, driverId, callback)
	if eventId == nil or callback == nil then
		Log:Error("Cannot schedule a system event without an event id or a callback")
		return
	end

	if not driverId then
		driverId = 0
	end

	if not self._callbacks[eventId] then
		self._callbacks[eventId] = {}
	end

	if not self._callbacks[eventId][driverId] then
		-- only register if we aren't already registered
		Log:Info("RegisterSystemEvent", eventId, driverId)
		C4:RegisterSystemEvent(eventId, driverId)
		self._callbacks[eventId][driverId] = {}
	else
		Log:Info("RegisterSystemEvent already registered for event:", eventId, "on driver:", driver)
	end

	table.insert(self._callbacks[eventId][driverId], callback)
end

function SystemEventManager:unregisterEvent(eventId, driverId, callback)
	print("SystemEventManager:unregisterEvent", eventId, driverId, callback)
	local event = self._callbacks[eventId]

	if event then
		print("found event")
		local driverCallbacks = event[driverId]
		if driverCallbacks then
			print(" found device", #driverCallbacks)
			for i,v in ipairs(driverCallbacks) do
				print("callback: ", v)
				if v == callback then
					table.remove(driverCallbacks, i)
					if # driverCallbacks <= 0 then
						C4:UnregisterSystemEvent(eventId, driverId)
						event[driverId] = nil
					end
					return 
				end
			end
		end
	end

	Log:Warn("Unable to unregister system event. Callback did not exist in event list")
end

--[[
	Called by onSystemEvent
--]]
function SystemEventManager:parseSystemEvent(data)
	local dataXml = C4:ParseXml(data)

	if "OnDataToUI" == dataXml.Attributes.name then
		self:_parseDataToUI(dataXml)
	else
		self:_fireCallback(C4SystemEvents[dataXml.Attributes.name], 0, dataXml)
	end
end

--[[
	Unregisters all events
--]]
function SystemEventManager:destroy()
	C4:UnregisterAllSystemEvents()
end

function SystemEventManager:_fireCallback(eventId, driverId, data)
	--print(eventId, driverId, data)
	if self._callbacks[eventId][driverId] then
		for i,callback in ipairs(self._callbacks[eventId][driverId]) do
			callback(data)
		end
	end
end

--[[
	Parses a data to ui system event. 
	@param data A DOMed XML object
--]]
function SystemEventManager:_parseDataToUI(data)
	local deviceId = 0
	local deviceData = nil

	for i,param in ipairs(data.ChildNodes) do
		if "iddevice" == param.Attributes.name then
			deviceId = param.Value
		elseif "data" == param.Attributes.name then
			deviceData = param.ChildNodes[1]
		end
	end

	self:_fireCallback(C4SystemEvents["OnDataToUI"], deviceId, deviceData)
end


