--------------------------------------------------------------------------------------------
--	driver.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--------------------------------------------------------------------------------------------

require "law.objects"

Utf8 = inheritsFrom(nil)

-- http://stackoverflow.com/questions/13235091/extract-the-first-letter-of-a-utf-8-string-with-lua
Utf8.pattern = "[%z\1-\127\194-\244][\128-\191]*"

--[[
	Creates the utf string
	@param s The lua string that has utf8
--]]
function Utf8:init(s)
	self._s = s

	-- Override the tostring method so that we 
	-- can do print(utf8Str) and it will print the 
	-- string self._s as is
	local mt = getmetatable(self)
	mt.__tostring = self.__tostring
end

function Utf8:__tostring()
	return self._s
end

--[[
	Returns the length of a utf8 encoded string
--]]
function Utf8:len()
	return select(2, self._s:gsub("[^\128-\193]",""))
end

--[[
	Creates an iterator that iterates over every character in the string.
	The fast iterator uses lua's pattern matching and should work for well
	formed utf8. It might match invalid characters if there is invalid utf8
--]]
function Utf8:fastIter()
	return self._s:gmatch("[%z\1-\127\194-\244][\128-\191]*")
end

--[[
	Uses the internal parsing methods to fully parse the utf8 document
	The iterator stops if an invalid utf8 character is found.
--]]
function Utf8:iter()
	local nextChar = 1
	return function()
			local str
			str, nextChar = self:_matchChar(nextChar)
			return str
	end
end

--[[
	Returns a lua string that represents the character
	at character index i. 

	Note: Do not use this method to iterate the string! It will 
	cause O(n^2) running time because each call itself 
	iterates over the string n times where n equals the index i 
	passed in to the function.
--]]
function Utf8:charAt(i)
	assert(i > 0, "Utf8:charAt invalid index i")

	-- get the iterator and call it
	local char = nil 
	local iterator = self:fastIter()
	local index = 0

	repeat
		char = iterator()
		index = index + 1
	until char == nil or index == i

	return char
end

--[[
	Returns a lua string that represents the utf8 char
	at the byte index i.
--]]
function Utf8:charAtByteIndex(i)
	if i < 1 or i > #self._s then 
		return nil
	end

	-- Walk backward at most 3 bytes to find
	-- a matching character
	for j = 0, 3 do
		local str = self:_matchChar(i - j)
		if str then
			return str
		end
	end

	return nil
end


--[[
	Matches a UTF8 1 byte character.
	@returns string, number. String is the matched character. number is the next character starting position. 
--]]
function Utf8:_match1(byteIndex)
	local byte = self:_byteAtIndex(byteIndex)

	if byte ~= nil and byte >= 0 and byte <= 0x7F then
		return string.char(byte), byteIndex + 1
	end

	return nil, nil
end

--[[
	Matches a UTF8 2 byte character.
	@returns string, number. String is the matched character. number is the next character starting position. 
--]]
function Utf8:_match2(byteIndex)
	local byte = self:_byteAtIndex(byteIndex)
	
	if byte == nil then
		return nil, nil
	end

	if byte >= 0xC2 and byte <= 0xDF and
	   self:_matchTailByIndex(byteIndex + 1) then
		return self._s:sub(byteIndex, byteIndex + 1), byteIndex + 2
	end

	return nil, nil
end

--[[
	Matches a UTF8 3 byte character.
	@returns string, number. String is the matched character. number is the next character starting position. 
--]]
function Utf8:_match3(byteIndex)
	local byte = self:_byteAtIndex(byteIndex)

	if byte == nil then
		return nil, nil
	end

	if byte == 0xE0 then
		byte = self:_byteAtIndex(byteIndex + 1)
		if 	byte >= 0xA0 and byte <= 0xBF and
		   	self:_matchTailByIndex(byteIndex + 2) then
			return self._s:sub(byteIndex, byteIndex + 2), byteIndex + 3			
		end

	elseif byte >= 0xE1 and byte <= 0xEC and
		   self:_matchTailByIndex(byteIndex + 1) and
		   self:_matchTailByIndex(byteIndex + 2) then
		   return self._s:sub(byteIndex, byteIndex + 2), byteIndex + 3

	elseif byte == 0xED then
		byte = self:_byteAtIndex(byteIndex + 1)
		if byte >= 0x80 and byte <= 0x9F and
		   self:_matchTailByIndex(byteIndex + 2) then
		   return self._s:sub(byteIndex, byteIndex + 2), byteIndex + 3
		end

	elseif byte >= 0xEE and byte <= 0xEF and
		   self:_matchTailByIndex(byteIndex + 1) and
		   self:_matchTailByIndex(byteIndex + 2) then
		   return self._s:sub(byteIndex, byteIndex + 2), byteIndex + 3
	end

	return nil, nil 
end

--[[
	Matches a UTF8 4 byte character.
	@returns string, number. String is the matched character. number is the next character starting position. 
--]]
function Utf8:_match4(byteIndex)
	local byte = self:_byteAtIndex(byteIndex)

	if byte == nil then
		return nil, nil
	end

	if byte == 0xF0 then
		byte = self:_byteAtIndex(byteIndex + 1)

		if byte >= 0x90 and byte <= 0xBF and
		   self:_matchTailByIndex(byteIndex + 2) and
		   self:_matchTailByIndex(byteIndex + 3) then
		   return self._s:sub(byteIndex, byteIndex + 3), byteIndex + 4
		end

	elseif byte >= 0xF1 and byte <= 0xF3 and
		   self:_matchTailByIndex(byteIndex + 1) and
		   self:_matchTailByIndex(byteIndex + 2) and
		   self:_matchTailByIndex(byteIndex + 3) then
		   return self._s:sub(byteIndex, byteIndex + 3), byteIndex + 4

	elseif byte == 0xF4 then
		byte = self:_byteAtIndex(byteIndex + 1)
		if byte >= 0x80 and byte <= 0x8F and
		   self:_matchTailByIndex(byteIndex + 2) and
		   self:_matchTailByIndex(byteIndex + 3) then
		   return self._s:sub(byteIndex, byteIndex + 3), byteIndex + 4
		  end
	end

	return nil, nil
end

--[[
	This will return the character iff the byteIndex starts on a
	character boundary. 
	@returns The matched character string
--]]
function Utf8:_matchChar(byteIndex)
	--if byteIndex < 1 or byteIndex >= #self._s then
	--	return nil, nil
	--end

	local str, nextIndex = self:_match1(byteIndex) 

	if not str then 
		str, nextIndex = self:_match2(byteIndex)
	end

	if not str then 
		str, nextIndex = self:_match3(byteIndex) 
	end

	if not str then 
		str, nextIndex = self:_match4(byteIndex) 
	end

	return str, nextIndex
end

function Utf8:_matchTail(byte)
	if byte >= 0x80 and byte <= 0xBF then
		return true
	end

	return false
end

function Utf8:_matchTailByIndex(byteIndex)
	return self:_matchTail(self:_byteAtIndex(byteIndex))
end

function Utf8:_byteAtIndex(index)
	return self._s:byte(index)
end

function Utf8.test()
	-- a = single byte
	-- Æ = 2 bytes
	-- Ø = 2 bytes
	-- Å = 2 bytes
	-- 😈 = 4 bytes
	-- ⓑ = 3 bytes

	local str = "ÆØÅa😈ⓑ"

	local utf8str = Utf8:new(str)
	
	assert(utf8str:len() == 6, "Invalid string len.")

	assert(utf8str:charAt(1) == "Æ", "CharAt(1) does not equal Æ")
	assert(utf8str:charAt(2) == "Ø", "CharAt(1) does not equal Ø")
	assert(utf8str:charAt(3) == "Å", "CharAt(1) does not equal Å")
	assert(utf8str:charAt(7) == nil, "CharAt(7) does not equal nil")

	assert(utf8str:_matchChar(1) == "Æ", [[matchChar(1) does not match "Æ"]])
	assert(utf8str:_matchChar(3) == "Ø", [[matchChar(3) does not match "Ø"]])
	assert(utf8str:_matchChar(5) == "Å", [[matchChar(5) does not match "Å"]])
	assert(utf8str:_matchChar(7) == "a", [[matchChar(7) does not match "a"]])
	assert(utf8str:_matchChar(8) == "😈", [[matchChar(8) does not match "😈"]])
	assert(utf8str:_matchChar(12) == "ⓑ", [[matchChar(12) does not match "ⓑ"]])

	assert(utf8str:charAtByteIndex(1) == "Æ", [[matchChar(1) does not match "Æ"]])
	assert(utf8str:charAtByteIndex(2) == "Æ", [[matchChar(1) does not match "Æ"]])

	assert(utf8str:charAtByteIndex(3) == "Ø", [[matchChar(3) does not match "Ø"]])
	assert(utf8str:charAtByteIndex(4) == "Ø", [[matchChar(4) does not match "Ø"]])

	assert(utf8str:charAtByteIndex(5) == "Å", [[matchChar(5) does not match "Å"]])
	assert(utf8str:charAtByteIndex(6) == "Å", [[matchChar(6) does not match "Å"]])

	assert(utf8str:charAtByteIndex(7) == "a", [[matchChar(7) does not match "a"]])
	
	assert(utf8str:charAtByteIndex(8) == "😈", [[matchChar(8) does not match "😈"]])
	assert(utf8str:charAtByteIndex(9) == "😈", [[matchChar(9) does not match "😈"]])
	assert(utf8str:charAtByteIndex(10) == "😈", [[matchChar(10) does not match "😈"]])
	assert(utf8str:charAtByteIndex(11) == "😈", [[matchChar(11) does not match "😈"]])

	assert(utf8str:charAtByteIndex(12) == "ⓑ", [[matchChar(12) does not match "ⓑ"]])
	assert(utf8str:charAtByteIndex(13) == "ⓑ", [[matchChar(12) does not match "ⓑ"]])
	assert(utf8str:charAtByteIndex(14) == "ⓑ", [[matchChar(12) does not match "ⓑ"]])

end

