--------------------------------------------------------------------------------------------
--	VariableManager.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Handles watching driver variables. Note in order for this class to function properly
--  it must be registered on the ServiceLocator as "VariableManager"
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.Log"

--[[
	C4 intrinsic method called when a registered variable changes
--]]
function OnWatchedVariableChanged(deviceId, variableId, strValue)
	ServiceLocator.VariableManager:OnWatchedVariableChanged(deviceId, variableId, strValue)
end


VariableManager = inheritsFrom(nil)

--[[
	Constructor
	@param registerServiceLocator Boolean indicates if the instance should automatically
			be added to the service locator
--]]
function VariableManager:init(registerServiceLocator)
	self._callbacks = {}

	if registerServiceLocator then
		ServiceLocator:addService("VariableManager", self)
	end
end

--[[
	Registers a specific callback to a variable
--]]
function VariableManager:registerVariable(deviceId, variableId, callback)
	Log:Debug("VariableManager:registerVariable", deviceId, variableId, callback)
	local deviceMap

	if self._callbacks[deviceId] then
		-- There is a map for the device id
		deviceMap = self._callbacks[deviceId]
	else
		-- Create a map for the device id.
		deviceMap = {}
		self._callbacks[deviceId] = deviceMap
	end

	local variableMap = deviceMap[variableId]

	if not variableMap then
		-- Create device variable map
		variableMap = {}
		deviceMap[variableId] = variableMap

		if not C4:RegisterVariableListener(deviceId, variableId) then
			deviceMap[variableId] = nil
			return
		end
	end

	-- Check if callback is already in the map
	for i, mapCallback in ipairs(variableMap) do
		if mapCallback == callback then
			return
		end
	end

	-- Whew we made it all the way. Add the callback to the variable map
	variableMap[#variableMap + 1] = callback
end

--[[
	unregisters a specific callback to a variable
--]]
function VariableManager:unregisterVariable(deviceId, variableId, callback)
	local deviceMap = self._callbacks[deviceId]

	if not deviceMap then
		return
	end

	local variableMap = deviceMap[variableId]

	if not variableMap then
		return
	end

	if #variableMap > 0 then

		local foundCallback
		for i, mapCallback in ipairs(variableMap) do
			if mapCallback == callback then
				table.remove(variableMap, i)
				break
			end
		end

		-- Was the last item removed?
		if #variableMap == 0 then
			deviceMap[variableId] = nil
			C4:UnregisterVariableListener(deviceId, variableId)
		end

	end
end

--[[
	Unregisters and clears all variables
--]]
function VariableManager:destroy()

	for deviceId, deviceMap in pairs(self._callbacks) do
		for variableId, variableMap in pairs(deviceMap) do
			C4:UnregisterVariableListener(deviceId, variableId)
		end
	end

	self._callbacks = {}
end

--[[
	Called by the system function. This object must be registerd on the ServiceLocator
	as "VariableManager" for this to work.
--]]
function VariableManager:OnWatchedVariableChanged(deviceId, variableId, strValue)
	-- Fire all of the callbacks registered for this variable
	local deviceMap = self._callbacks[deviceId]

	if not deviceMap then
		Log:Warn("VariableManager: No variable listener is registered for:", deviceId)
		return
	end

	local variableMap = deviceMap[variableId]

	if not variableMap then
		Log:Warn("VariableManager: No callback is registered for:", deviceId, variableId)
		return
	end

	for i, callback in ipairs(variableMap) do
		-- Fire the callback.
		callback(strValue)
	end
end
