--------------------------------------------------------------------------------------------
--	XML.lua
--  
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Helpers for dealing with XML
--------------------------------------------------------------------------------------------

require "law.objects"
require "law.StringUtils"

Xml = inheritsFrom(nil)


--------------------------------------------------------------------------------------------
--										STATIC METHODS	 									  --
--------------------------------------------------------------------------------------------

--[[
	Converts a table to xml. 
	@param tag The root tag to add.
	@param tData The data table to convert
	@param escapeValue if items are supposed to be escaped
	@param rawItemNames An array of property names that contain raw xml. When an item with one of these
			names is encountered the value will be injected directly into the string.
--]]
function Xml.BuildSimpleXml(tag, tData, escapeValue, rawItemNames)
	local xml = StringBuilder:new()
	
	-- Add a root tag if defined.
	if (tag ~= nil) then
		xml:add("<")
		xml:add(tag)
		xml:add(">")
	end
	
	local isRaw = false
	
	-- Data is in side of a flat table i.e. not a table of a table.
	for i,v in pairs(tData or {}) do
		isRaw = false
		for _, rawName in ipairs(rawItemNames or {}) do
			if rawName == i then
				isRaw = true
				xml:add(v)
			end
		end

		if not isRaw then
			xml:add("<")
			xml:add(i)
			xml:add(">")
			
			if escapeValue == true then
				xml:add(C4:XmlEscapeString(tostring(v)))
			else
				xml:add(v)
			end

			xml:add("</")
			xml:add(i)
			xml:add(">")
		end
	end
	
	-- End the root tag if defined.
	if (tag ~= nil) then
		xml:add("</")
		xml:add(tag)
		xml:add(">")
	end
	-- Get the string out of the string builder
	return tostring(xml)
end

--------------------------------------------------------------------------------------------
--										END STATIC METHODS	 									  --
--------------------------------------------------------------------------------------------

--------------------------------------------------------------------------------------------
--										INITIALIZATION	 									  --
--------------------------------------------------------------------------------------------

--[[
	Initializes an xml object.
	@param  Either an xml str or a C4 xml type
--]]
function Xml:init(xml)
	if type(xml) == "string" then
		self._xml = C4:ParseXml(xml)
	else
		self._xml = xml
	end

	local mt = getmetatable(self)
	assert(mt, "Xml no metatable on self")
	-- Save our current metatable 
	setmetatable(self._xml, mt)

	-- Allow this new class to index anything in C4 xml class
	setmetatable(self, {__index = self._xml})
end

--------------------------------------------------------------------------------------------
--										END INITIALIZATION	 									  --
--------------------------------------------------------------------------------------------

--------------------------------------------------------------------------------------------
--										METHODS	 									  --
--------------------------------------------------------------------------------------------


--[[
	Finds a node in the xml object.
	@param xml The xml (Must be same type as returned by C4:ParseXml)
	@param nodeName The name of the node to find
	@return The node that was found with name equal to nodeName or nil
--]]
function Xml:findNode(nodeName)
	for i,v in ipairs(self.ChildNodes) do
		if v.Name == nodeName then
			return Xml:new(v)
		end
	end

	return nil
end

--[[
	Add new child to current children
--]]
function Xml:addChildNode(node)
	table.insert(self.ChildNodes, node._xml)
end


--[[
	Iterates over all nodes with the given name
--]]
function Xml:nodesWithName(nodeName)
	local index = 1
	return function()
		for i=index, #self.ChildNodes do
			if self.ChildNodes[i].Name == nodeName then
				index = i + 1
				return Xml:new(self.ChildNodes[i])
			end
		end

		return nil
	end
end

function Xml:children() 
	local i = 0
    
    return function ()
    	i = i + 1
        if i <= #self.ChildNodes then 
        	return Xml:new(self.ChildNodes[i])
        end
    end
end

function Xml:findNodeWithPath(path)
	local parts = path:split("/")

	if #parts <= 0 then
		return nil
	elseif parts[1] == "" then
		-- An empty path returns a single empty path element. 
		-- This should is the identity path i.e. it is equal to itself. 
		return self
	end

	local result = self

	for i, pathPart in ipairs(parts) do
		result = result:findNode(pathPart)

		if not result then
			return nil
		end
	end

	return result

end

--------------------------------------------------------------------------------------------
--										END METHODS	 									  --
--------------------------------------------------------------------------------------------

