--------------------------------------------------------------------------------------------
--  Copyright 2014 Control4 Corporation. All rights reserved.
--
--  Defines base object framework and various helper methods.
--------------------------------------------------------------------------------------------

function inheritsFrom( baseClass )

    local new_class = {}
    local class_mt = { __index = new_class }

    function new_class:create(...)
        local newinst = {}
        setmetatable( newinst, class_mt )
        -- Call the constructor when we create this class
        if newinst.init then
            -- Allow returning a different obj than self. This allows for readonly tables etc...
            newinst = newinst:init(unpack({...})) or newinst
        end
        return newinst
    end

    if nil ~= baseClass then
        setmetatable( new_class, { __index = baseClass } )
    end

    -- Implementation of additional OO properties starts here --

    -- Return the class object of the instance
    function new_class:class()
        return new_class
    end

    --[[
        Return the super class object of the instance.
        @Note Calling methods on the base class itself will modify
            the base table's static properties. In order to have call
            the base class' methods and have them modify the current object
            use super() or superAsSelf().
    --]]
    function new_class:superClass()
        return baseClass
    end

    --[[
        Returns a table that allows calling of the base class's method
        while maintaining the objects state as the modified state of the base
        class' methods. For example consider the following statements (order matters):
            -- The child sees the parents property if the child hasn't overriden the property
            obj:superClass().id = "parent"
            obj.id == "parent" -- true

            -- Setting the property on the child overrides (hides) the parents property
            obj.id = "child"
            obj.id == "child" -- true

            -- The super() method pass
            obj:super().id == "parent" -- true
            obj:super().id = "child"
            obj:super().id == "parent" -- still true
            obj.id == "child" -- still true
    --]]
    function new_class:super()
        local holder = {}
        holder.child = self
        holder.parent = baseClass

        local mt = {}
        mt.__index = function(table, index)
            if table.parent[index] then
                return table.parent[index]
            else
                return table.child[index]
            end
        end

        -- Only set the new values to the child.
        mt.__newindex = function(table, key, value)
            table.child[key] = value
        end

        mt.__tostring = function(table)
            return tostring(table.child)
        end

        setmetatable(holder, mt)
        return holder
    end

    new_class.new = new_class.create

    -- Return true if the caller is an instance of theClass
    function new_class:isa( theClass )
        local b_isa = false

        local cur_class = new_class

        while ( nil ~= cur_class ) and ( false == b_isa ) do
            if cur_class == theClass then
                b_isa = true
            else
                cur_class = cur_class:superClass()
            end
        end

        return b_isa
    end

    return new_class
end

--[[
    Inheritance units tests
--]]
function __test_inheritance()
    local b = inheritsFrom(nil)
    b.init = function(self, msg)
        self._msg = msg
        self._type = "b"
    end

    local t = inheritsFrom(b)
    t.init = function(self, msg)
        self:super():init(msg)
       -- self._type = "t"
    end

    local v = inheritsFrom(t)
    function v:init(msg)
       self:super():init(msg)
       self._type = "v"
    end

    t1 = t:new("t1")
    t2 = t:new("t2")
    assert(t1._msg == "t1", "t1 message is not equal to 't1' it's: " .. t1._msg)
    assert(t2._msg == "t2", "t2 message is not equal to 't2' it's: " .. t2._msg)
    assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");

    assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

    t1:superClass().id = "parent"
    assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

    -- Setting the property on the child overrides (hides) the parents property
    t1.id = "child"
    assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

    -- The super() method maintains the self pointer to the child and not to the base
    assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
    t1:super().id = "child1"
    assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
    assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")

    v1 = v:new("v1")
    v2 = v:new("v2")
    assert(v1._msg == "v1", "v1 message is not equal to 'v1' it's: " .. v1._msg)
    assert(v2._msg == "v2", "v2 message is not equal to 'v2' it's: " .. v2._msg)
    assert(tostring(v1:super()) ==  tostring(v1), "v1 tostrings don't match");
    assert(v1._type == "v", "v1 type is not v. It's: " .. v1._type)
    assert(v1:super()._type == "v", "v1 super type ~= v it's: " .. v1:super()._type )
    assert(t._type == nil, "t has _type when it shouldn't")
    assert(b._type == nil, "b has _type when it shouldn't")

end

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function tableConcat(t1, t2)
    for i=1,#t2 do
        t1[#t1+1] = t2[i]
    end
    return t1
end

--[[
    Protects tables by making them readonly
--]]
function makeTableReadOnly(table)
   return setmetatable({}, {
     __index = table,
     __newindex = function(table, key, value)
                    error("Attempt to modify read-only table")
                  end,
     __metatable = false
   });
end

--[[
    Takes a vararg list and converts it to a table
--]]
function pack(...)
    return {...}
end

function pcallAndPrint(func, ...)
    local result = pack(pcall(func, ...))
    if result[1] == false then
        print("pcall error: ", result[2])
        if debug then
            print(debug.traceback())
        end
    end
    return unpack(result)
end

function toboolean(val)
    local valType = type(val)

    if "boolean" == valType then
        return val

    elseif val == nil then
        return false

    elseif "string" == valType then

        local numVal = tonumber(val)

        if numVal == nil then
            -- match string
            if val:lower() == "false" then
                return false
            end
        else
            val = numVal
        end
    end

    if val ~= 0 then
        return true
    end

    return false
end

function os.capture(cmd, raw)
  local f = assert(io.popen(cmd, 'r'))
  local s = assert(f:read('*a'))
  f:close()
  if raw then return s end
  s = string.gsub(s, '^%s+', '')
  s = string.gsub(s, '%s+$', '')
  s = string.gsub(s, '[\n\r]+', ' ')
  return s
end
